from __future__ import annotations
from typing import Union, Optional, Literal

import numpy as np
import pandas as pd
from scipy.optimize import minimize

from .dataprep import Dataprep
from .base import BaseSynth, VanillaOptimMixin
from .inference import ConformalInference


OptimizerMethod_t = Literal[
    "Nelder-Mead", "Powell", "CG", "BFGS", "L-BFGS-B", "TNC", "COBYLA", "trust-constr"
]


class Synth(BaseSynth, VanillaOptimMixin):
    """Implementation of the synthetic control method due to
    Abadie & Gardeazabal :cite:`basque2003`.
    """

    def __init__(self) -> None:
        super().__init__()
        self.loss_W: Optional[float] = None
        self.V: Optional[np.ndarray] = None
        self.loss_V: Optional[float] = None

    def fit(
        self,
        dataprep: Optional[Dataprep] = None,
        X0: Optional[pd.DataFrame] = None,
        X1: Optional[pd.Series] = None,
        Z0: Optional[pd.DataFrame] = None,
        Z1: Optional[pd.Series] = None,
        custom_V: Optional[np.ndarray] = None,
        optim_method: OptimizerMethod_t = "Nelder-Mead",
        optim_initial: Literal["equal", "ols"] = "equal",
        optim_options: dict = {"maxiter": 1000},
    ) -> None:
        """Fit the model/calculate the weights. Either a :class:`Dataprep` object
        should be provided or otherwise matrices (:math:`X_0`, :math:`X_1`, :math:`Z_0`,
        :math:`Z_1`) should be provided (using the notation of Abadie &
        Gardeazabal :cite:`basque2003`).

        Parameters
        ----------
        dataprep : Dataprep, optional
            :class:`Dataprep` object containing data to model, by default None.
        X0 : pd.DataFrame, shape (m, c), optional
            Matrix with each column corresponding to a control unit and each
            row is covariates, by default None.
        X1 : pandas.Series, shape (m, 1), optional
            Column vector giving the covariate values for the treated unit, by
            default None.
        Z0 : pandas.DataFrame, shape (n, c), optional
            A matrix of the time series of the outcome variable with each
            column corresponding to a control unit and the rows are the time
            steps; the columns correspond with the columns of X0, by default
            None.
        Z1 : pandas.Series, shape (n, 1), optional
            Column vector giving the outcome variable values over time for the
            treated unit, by default None.
        custom_V : numpy.ndarray, shape (c, c), optional
            Provide a V matrix (using the notation of the Abadie, Diamond &
            Hainmueller paper), the optimisation problem will only then be
            solved for the weight matrix W, by default None.
        optim_method : str, optional
            Optimisation method to use for the outer optimisation, can be
            any of the valid options for scipy minimize that do not require a
            jacobian matrix, namely

                - 'Nelder-Mead'
                - 'Powell'
                - 'CG'
                - 'BFGS'
                - 'L-BFGS-B'
                - 'TNC'
                - 'COBYLA'
                - 'trust-constr'

            By default 'Nelder-Mead'.
        optim_initial : str, optional
            Starting value for the outer optimisation, possible starting
            values are

                - 'equal', where the weights are all equal,
                - 'ols', which uses a starting value obtained for fitting a
                  regression.

            By default 'equal'.
        optim_options : dict, optional
            options to provide to the outer part of the optimisation, value
            options are any option that can be provided to scipy minimize for
            the given optimisation method, by default `{'maxiter': 1000}`.

        Returns
        -------
        NoneType
            None

        Raises
        ------
        ValueError
            if neither a `Dataprep` object nor all of (:math:`X_0`, :math:`X_1`,
            :math:`Z_0`, :math:`Z_1`) are supplied.
        TypeError
            if (:math:`X1`, :math:`Z1`) are not of type `pandas.Series`.
        ValueError
            if `optim_initial=ols` and there is collinearity in the data.
        ValueError
            if `optim_initial` is not one of `'equal'` or `'ols'`.
        """
        if dataprep:
            if (
                isinstance(dataprep.treatment_identifier, (list, tuple))
                and len(dataprep.treatment_identifier) > 1
            ):
                raise ValueError("Synth requires exactly one treated unit.")
            self.dataprep = dataprep
            X0, X1 = dataprep.make_covariate_mats()
            Z0, Z1 = dataprep.make_outcome_mats()
        else:
            if X0 is None or X1 is None or Z0 is None or Z1 is None:
                raise ValueError(
                    "dataprep must be set or (X0, X1, Z0, Z1) must all be set."
                )
            if not isinstance(X1, pd.Series) or not isinstance(Z1, pd.Series):
                raise TypeError("X1 and Z1 must be of type `pandas.Series`.")

        X = pd.concat([X0, X1], axis=1)
        X_scaled = X.divide(X.std(axis=1), axis=0)
        X0_scaled, X1_scaled = X_scaled.drop(columns=X1.name), X_scaled[X1.name]

        X0_arr = X0_scaled.to_numpy()
        X1_arr = X1_scaled.to_numpy()
        Z0_arr = Z0.to_numpy()
        Z1_arr = Z1.to_numpy()

        if custom_V is not None:
            V_mat = np.diag(custom_V)
            W, loss_W = self.w_optimize(V_mat=V_mat, X0=X0_arr, X1=X1_arr)
            loss_V = self.calc_loss_V(W=W, Z0=Z0_arr, Z1=Z1_arr)
            self.W, self.loss_W, self.V, self.loss_V = W, loss_W, custom_V, loss_V
            return

        n_r, _ = X0_arr.shape

        if optim_initial == "equal":
            x0 = [1 / n_r] * n_r
        elif optim_initial == "ols":
            X_arr = np.hstack([X0_arr, X1_arr.reshape(-1, 1)])
            X_arr = np.hstack([np.full((X_arr.shape[1], 1), 1), X_arr.T])
            Z_arr = np.hstack([Z0_arr, Z1_arr.reshape(-1, 1)])

            try:
                beta = np.linalg.inv(X_arr.T @ X_arr) @ X_arr.T @ Z_arr.T
            except np.linalg.LinAlgError:
                raise ValueError(
                    'Could not invert X^T.X required for `optim_initial="ols"`, '
                    "probably there is collinearity in your data."
                )

            beta = beta[1:,]  # fmt: skip
            x0 = np.diag(beta @ beta.T)
            x0 = x0 / sum(x0)
        else:
            raise ValueError("Unknown option for `optim_initial`.")

        def fun(x):
            V_mat = np.diag(np.abs(x)) / np.sum(np.abs(x))
            W, _ = self.w_optimize(V_mat=V_mat, X0=X0_arr, X1=X1_arr)
            loss_V = self.calc_loss_V(W=W, Z0=Z0_arr, Z1=Z1_arr)
            return loss_V

        res = minimize(fun=fun, x0=x0, method=optim_method, options=optim_options)
        V_mat = np.diag(np.abs(res["x"])) / np.sum(np.abs(res["x"]))
        W, loss_W = self.w_optimize(V_mat=V_mat, X0=X0_arr, X1=X1_arr)
        loss_V = self.calc_loss_V(W=W, Z0=Z0_arr, Z1=Z1_arr)

        self.W, self.loss_W, self.V, self.loss_V = W, loss_W, V_mat.diagonal(), loss_V
        self.W_names = Z0.columns

    @staticmethod
    def calc_loss_V(W: np.ndarray, Z0: np.ndarray, Z1: np.ndarray) -> float:
        """Calculates the V loss.

        Parameters
        ----------
        W : numpy.ndarray, shape (n,)
            Vector of the control weights
        Z0 : numpy.ndarray, shape (m, n)
            Matrix of the time series of the outcome variable with each
            column corresponding to a control unit and the rows are the time
            steps.
        Z1 : numpy.ndarray, shape (m,)
            Column vector giving the outcome variable values over time for the
            treated unit

        Returns
        -------
        float
            V loss.

        :meta private:
        """
        loss_V = (Z1 - Z0 @ W).T @ (Z1 - Z0 @ W) / len(Z0)
        return loss_V.item()

    def summary(
        self,
        round: int = 3,
        X0: Optional[pd.DataFrame] = None,
        X1: Optional[pd.Series] = None,
    ) -> pd.DataFrame:
        """Generates a ``pandas.DataFrame`` with summary data. In particular,
        it will show the values of the V matrix for each predictor, then the
        next column will show the mean value of each predictor over the time
        period ``time_predictors_prior`` for the treated unit and the synthetic
        unit and finally there will be a column 'sample mean' that shows the
        mean value of each predictor over the time period
        ``time_predictors_prior`` across all the control units, i.e. this will
        be the same as a synthetic control where all the weights are equal.

        Parameters
        ----------
        round : int, optional
            Round the numbers to given number of places, by default 3
        X0 : pd.DataFrame, shape (n_cov, n_controls), optional
            Matrix with each column corresponding to a control unit and each
            row is a covariate. If no dataprep is set, then this must be
            supplied along with X1, by default None.
        X1 : pandas.Series, shape (n_cov, 1), optional
            Column vector giving the covariate values for the treated unit.
            If no dataprep is set, then this must be supplied along with Z1,
            by default None.

        Returns
        -------
        pandas.DataFrame
            Summary data.

        Raises
        ------
        ValueError
            If there is no V matrix available
        ValueError
            If there is no :class:`Dataprep` object set or (Z0, Z1) is not supplied
        ValueError
            If there is no weight matrix available
        """
        if self.V is None:
            raise ValueError("No V matrix available; fit data first.")
        summary_ser = super().summary(round=round, X0=X0, X1=X1)

        V = pd.Series(self.V, index=summary_ser.index, name="V")
        return pd.concat([V, summary_ser], axis=1).round(round)

    def confidence_interval(
        self,
        alpha: float,
        time_periods: list,
        tol: float,
        pre_periods: Optional[list] = None,
        dataprep: Optional[Dataprep] = None,
        X0: Optional[pd.DataFrame] = None,
        X1: Optional[pd.Series] = None,
        Z0: Optional[pd.DataFrame] = None,
        Z1: Optional[pd.Series] = None,
        custom_V: Optional[np.ndarray] = None,
        optim_method: OptimizerMethod_t = None,
        optim_initial: Literal["equal", "ols"] = None,
        optim_options: dict = None,
        method: Literal["conformal"] = "conformal",
        max_iter: int = 50,
        step_sz: Optional[float] = None,
        step_sz_div: float = 20.0,
        verbose: bool = True,
    ) -> pd.DataFrame:
        """Confidence intervals obtained from test-inversion, where
        the p-values are obtained by adjusted refits of the data
        following Chernozhukov et al. :cite:`inference2021`.

        Parameters
        ----------
        alpha : float
            The required significance level, e.g. alpha = 0.05 will
            yield a confidence level of 100 * (1 - alpha) = 95%.
        time_periods : list
            The time-periods to calculate confidence intervals for.
        tol : float
            The required tolerance (accuracy) required when calculating the
            lower/upper cut-off point of the confidence interval. The search
            will try to obtain this tolerance level but will not exceed `max_iter`
            iterations trying to achieve that.
        pre_periods : Optional[list], optional
            The time-periods to use for the optimization when refitting the
            data with the adjusted outcomes, optional.
        dataprep : Optional[Dataprep], optional
            Dataprep object defining the study data, if this is not supplied
            then either self.dataprep must be set or else (X0, X1, Z0, Z1) must
            all be supplied, by default None.
        X0 : pd.DataFrame, shape (m, c), optional
            Matrix with each column corresponding to a control unit and each
            row is covariates, if this is not supplied then either `dataprep` must
             be supplied or `self.dataprep` must be set by default None.
        X1 : pandas.Series, shape (m, 1), optional
            Column vector giving the covariate values for the treated unit, if
            this is not supplied then either `dataprep` must
             be supplied or `self.dataprep` must be set by default None.
        Z0 : pandas.DataFrame, shape (n, c), optional
            A matrix of the time series of the outcome variable with each
            column corresponding to a control unit and the rows are the time
            steps; the columns correspond with the columns of X0, if this
            is not supplied then either `dataprep` must be supplied or
            `self.dataprep` must be set by default None.
        Z1 : pandas.Series, shape (n, 1), optional
            Column vector giving the outcome variable values over time for the
            treated unit, if this is not supplied then either `dataprep` must
             be supplied or `self.dataprep` must be set by default None.
        custom_V : numpy.ndarray, shape (c, c), optional
            Provide a V matrix (using the notation of the Abadie, Diamond &
            Hainmueller paper), the optimisation problem will only then be
            solved for the weight matrix W. This is the same argument
            as in the `fit` method, by default None.
        optim_method : str, optional
            Optimisation method to use for the outer optimisation, can be
            any of the valid options for scipy minimize that do not require a
            jacobian matrix, namely

                - 'Nelder-Mead'
                - 'Powell'
                - 'CG'
                - 'BFGS'
                - 'L-BFGS-B'
                - 'TNC'
                - 'COBYLA'
                - 'trust-constr'

            This is the same argument as in the `fit` method, by default
            'Nelder-Mead'.
        optim_initial : str, optional
            Starting value for the outer optimisation, possible starting
            values are

                - 'equal', where the weights are all equal,
                - 'ols', which uses a starting value obtained for fitting a
                  regression.

            This is the same argument as in the `fit` method, by default
            'equal'.
        optim_options : dict, optional
            options to provide to the outer part of the optimisation, value
            options are any option that can be provided to scipy minimize for
            the given optimisation method. This is the same argument as in
             the `fit` method, by default `{'maxiter': 1000}`.
        method : str, optional
            The type of method to use when computing the confidence intervals,
            currently only conformal inference (`conformal`) is implemented,
            by default "conformal".
        max_iter : int, optional
            Maximum number of times to re-fit the data when trying to locate
            the lower/upper cut-off point and when binary searching for the
            cut-off point, by default 20.
        step_sz : Optional[float], optional
            Step size to use when searching for an interval that contains the
            lower or upper cut-off point of the confidence interval, by default None.
        step_sz_div : float, optional
            Alternative way to define step size: it is the fraction that defines
            step-size in terms of the standard deviation of the att, i.e. if
            `step_sz_div=20.0` then the step size used will be (att +/- 2.5 * std(att)) / 20.0,
            by default 20.0.
        verbose : bool, optional
            Print output, by default True.

        Returns
        -------
        pd.DataFrame
            A pandas.DataFrame indexed by `post_periods`, with 3 columns: `value` that
            gives the calculated treatment effect, `lower_ci` that gives the value
            defining the lower-end of the confidence interval, `upper_ci` that gives
            the value defining the upper-end of the confidence interval.

        Raises
        ------
        ValueError
            If there is no :class:`Dataprep` object set or (X0, X1, Z0, Z1) is not supplied or
            `self.dataprep` is not set.
        TypeError
            if (:math:`X1`, :math:`Z1`) are not of type `pandas.Series`.
        ValueError
            if `dataprep` is not set and `pre-periods` is not set.
        ValueError
            if an invalid option for `method` is given, currently only `conformal` is supported.
        """
        if method == "conformal":
            if dataprep is not None:
                X0, X1 = dataprep.make_covariate_mats()
                if pre_periods is None:
                    pre_periods = list(dataprep.time_optimize_ssr)
                if 1.0 / len(pre_periods) > alpha:
                    raise ValueError(
                        "Too few pre-intervention time-periods available for "
                        f"significance level `alpha`={alpha}, either increase `alpha` "
                        "or use more pre-intervention time-periods."
                    )
                all_time_periods = time_periods + list(pre_periods)
                Z0, Z1 = dataprep.make_outcome_mats(time_period=all_time_periods)
            elif self.dataprep is not None:
                X0, X1 = self.dataprep.make_covariate_mats()
                if pre_periods is None:
                    pre_periods = list(self.dataprep.time_optimize_ssr)
                if 1.0 / len(pre_periods) > alpha:
                    raise ValueError(
                        "Too few pre-intervention time-periods available for "
                        f"significance level `alpha`={alpha}, either increase `alpha` "
                        "or use more pre-intervention time-periods."
                    )
                all_time_periods = time_periods + list(pre_periods)
                Z0, Z1 = self.dataprep.make_outcome_mats(time_period=all_time_periods)
            else:
                if X0 is None or X1 is None or Z0 is None or Z1 is None:
                    raise ValueError(
                        "dataprep must be set or (X0, X1, Z0, Z1) must all be set."
                    )
                if not isinstance(X1, pd.Series) or not isinstance(Z1, pd.Series):
                    raise TypeError("X1 and Z1 must be of type `pandas.Series`.")
                if pre_periods is None:
                    raise ValueError("`pre_periods` must be set if not using dataprep.")
                if 1.0 / len(pre_periods) > alpha:
                    raise ValueError(
                        "Too few pre-intervention time-periods available for "
                        f"significance level `alpha`={alpha}, either increase `alpha` "
                        "or use more pre-intervention time-periods."
                    )

            scm_fit_args = {"X0": X0, "X1": X1}
            if custom_V is not None:
                scm_fit_args["custom_V"] = custom_V
            if optim_method:
                scm_fit_args["optim_method"] = optim_method
            if optim_initial:
                scm_fit_args["optim_initial"] = optim_initial
            if optim_options:
                scm_fit_args["optim_options"] = optim_options

            conformal_inf = ConformalInference()
            df_cis = conformal_inf.confidence_intervals(
                alpha=alpha,
                scm=self,
                Z0=Z0,
                Z1=Z1,
                pre_periods=pre_periods,
                post_periods=time_periods,
                scm_fit_args=scm_fit_args,
                max_iter=max_iter,
                tol=tol,
                step_sz=step_sz,
                step_sz_div=step_sz_div,
                verbose=verbose,
            )
            return df_cis
        else:
            raise ValueError("Invalid option for `method`.")



from __future__ import annotations
from typing import Optional

import numpy as np

from .dataprep import Dataprep
from .base import BaseSynth


class RobustSynth(BaseSynth):
    """Implementation of the robust synthetic control method due to
    Amjad, Shah & Shen :cite:`robust2018`.
    """

    def __init__(self) -> None:
        super().__init__()
        self.W: Optional[np.ndarray] = None
        self.lambda_: Optional[float] = None

    def fit(
        self,
        dataprep: Dataprep,
        lambda_: float,
        threshold: Optional[float] = None,
        sv_count: Optional[int] = None,
    ) -> None:
        """Fit the model/calculate the weights.

        Parameters
        ----------
        dataprep : Dataprep
            :class:`Dataprep` object containing data to model.
        lambda_ : float
            Ridge parameter to use.
        threshold: float, optional
            Remove singular values that are less than this threshold.
        sv_count: int, optional
            Keep this many of the largest singular values when
            reducing the outcome matrix
        """
        if (
            isinstance(dataprep.treatment_identifier, (list, tuple))
            and len(dataprep.treatment_identifier) > 1
        ):
            raise ValueError("RobustSynth requires exactly one treated unit.")
        self.dataprep = dataprep
        time_period_min = dataprep.foo[dataprep.time_variable].astype("int").min()
        time_period_max = dataprep.foo[dataprep.time_variable].astype("int").max()

        X0, X1 = dataprep.make_outcome_mats(
            time_period=range(time_period_min, 1 + time_period_max)
        )
        Y = X0.T.values

        M_hat = self._sv_decomposition(Y, threshold, sv_count).T

        time_optim_end = 1 + dataprep.time_optimize_ssr[-1]
        end_idx = X0.index.to_list().index(time_optim_end)
        M_hat_neg = M_hat[:end_idx, :]
        Y1_neg = X1.to_numpy()[:end_idx]

        self.W = np.matmul(
            np.linalg.inv(
                M_hat_neg.T @ M_hat_neg + lambda_ * np.identity(M_hat_neg.shape[1])
            ),
            M_hat_neg.T @ Y1_neg,
        )

    def _sv_decomposition(
        self,
        Y: np.ndarray,
        threshold: Optional[float] = None,
        sv_count: Optional[int] = None,
    ) -> np.ndarray:
        """Calculate the :math:`\hat{M}` matrix from the paper (see :cite:`robust2018`) by
        carrying out an SVD of the outcome matrix and remove the specified number
        of singular values.

        Parameters
        ----------
        Y : np.ndarray
            The outcome matrix (:math:`Y` matrix in the notation of the paper).
        threshold : Optional[float], optional
            Remove singular values that are less that `threshold`,
            either this must be specified or `sv_count`, by default None
        sv_count : Optional[int], optional
            Keep this many of the largest singular values,
            either this must be specified or `threshold`, by default None

        Returns
        -------
        np.ndarray
            The :math:`\hat{M}` matrix from the paper (see :cite:`robust2018`).

        Raises
        ------
        ValueError
            If neither `threshold` nor `sv_count` are supplied.

        :meta private:
        """
        if not threshold and not sv_count:
            raise ValueError("One of `threshold` or `sv_count` must be supplied.")

        if threshold:
            idx = 0
            while s[idx] > threshold:
                idx += 1
        else:
            idx = sv_count

        u, s, v = np.linalg.svd(Y)
        s_res = np.zeros_like(Y)
        s_res[:idx, :idx] = np.diag(s[:idx])

        r, c = Y.shape
        p_hat = max(np.count_nonzero(Y) / (r * c), 1 / (r * c))
        M_hat = (1 / p_hat) * (u @ s_res @ v)
        return M_hat



import unittest
import numpy as np
import pandas as pd

from pysyncon import Dataprep
from pysyncon.dataprep import AGG_OP


class TestDataprep(unittest.TestCase):
    def setUp(self):
        # 1 -> treated, (2, 3) -> controls
        self.foo = pd.DataFrame(
            {
                "time": [1, 2, 3, 4, 1, 2, 3, 4, 1, 2, 3, 4],
                "name": [1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3],
                "dependent": np.random.random(12),
                "predictor1": np.random.random(12),
                "predictor2": np.random.random(12),
            }
        )
        self.predictors = ["predictor1"]
        self.predictors_op = "mean"
        self.dependent = "dependent"
        self.unit_variable = "name"
        self.time_variable = "time"
        self.treatment_identifier = 1
        self.treatment_identifier_list = [1, 2]
        self.controls_identifier = [2, 3]
        self.controls_identifier_alt = [3]
        self.time_predictors_prior = [2, 3]
        self.time_optimize_ssr = [1, 2, 3]
        self.special_predictors = [
            ("predictor1", [2], "mean"),
            ("predictor2", [1, 2], "median"),
            ("predictor2", [1, 2], "std"),
        ]
        self.str_no_special = (
            "Dataprep\n"
            "Treated unit: 1\n"
            "Dependent variable: dependent\n"
            "Control units: 2, 3\n"
            "Time range in data: 1 - 4\n"
            "Time range for loss minimization: [1, 2, 3]\n"
            "Time range for predictors: [2, 3]\n"
            "Predictors: predictor1\n"
        )
        self.str_special = (
            "Dataprep\n"
            "Treated unit: 1\n"
            "Dependent variable: dependent\n"
            "Control units: 2, 3\n"
            "Time range in data: 1 - 4\n"
            "Time range for loss minimization: [1, 2, 3]\n"
            "Time range for predictors: [2, 3]\n"
            "Predictors: predictor1\n"
            "Special predictors:\n"
            "    `predictor1` over `[2]` using `mean`\n"
            "    `predictor2` over `[1, 2]` using `median`\n"
            "    `predictor2` over `[1, 2]` using `std`\n"
        )

    def test_init_arg_foo(self):
        kwargs = {
            "predictors": self.predictors,
            "predictors_op": self.predictors_op,
            "dependent": self.dependent,
            "unit_variable": self.unit_variable,
            "time_variable": self.time_variable,
            "treatment_identifier": self.treatment_identifier,
            "controls_identifier": self.controls_identifier,
            "time_predictors_prior": self.time_predictors_prior,
            "time_optimize_ssr": self.time_optimize_ssr,
            "special_predictors": self.special_predictors,
        }

        self.assertRaises(TypeError, Dataprep, foo=np.array([]), **kwargs)
        self.assertRaises(TypeError, Dataprep, foo=list(), **kwargs)
        self.assertRaises(TypeError, Dataprep, foo=tuple(), **kwargs)

    def test_init_arg_predictors(self):
        kwargs = {
            "foo": self.foo,
            "predictors_op": self.predictors_op,
            "dependent": self.dependent,
            "unit_variable": self.unit_variable,
            "time_variable": self.time_variable,
            "treatment_identifier": self.treatment_identifier,
            "controls_identifier": self.controls_identifier,
            "time_predictors_prior": self.time_predictors_prior,
            "time_optimize_ssr": self.time_optimize_ssr,
            "special_predictors": self.special_predictors,
        }

        self.assertRaises(ValueError, Dataprep, predictors=["badval"], **kwargs)

    def test_init_arg_predictors_op(self):
        kwargs = {
            "foo": self.foo,
            "predictors": self.predictors,
            "dependent": self.dependent,
            "unit_variable": self.unit_variable,
            "time_variable": self.time_variable,
            "treatment_identifier": self.treatment_identifier,
            "controls_identifier": self.controls_identifier,
            "time_predictors_prior": self.time_predictors_prior,
            "time_optimize_ssr": self.time_optimize_ssr,
            "special_predictors": self.special_predictors,
        }

        self.assertRaises(ValueError, Dataprep, predictors_op="badval", **kwargs)

    def test_init_arg_dependent(self):
        kwargs = {
            "foo": self.foo,
            "predictors": self.predictors,
            "predictors_op": self.predictors_op,
            "unit_variable": self.unit_variable,
            "time_variable": self.time_variable,
            "treatment_identifier": self.treatment_identifier,
            "controls_identifier": self.controls_identifier,
            "time_predictors_prior": self.time_predictors_prior,
            "time_optimize_ssr": self.time_optimize_ssr,
            "special_predictors": self.special_predictors,
        }

        self.assertRaises(ValueError, Dataprep, dependent="badval", **kwargs)

    def test_init_arg_unit_variable(self):
        kwargs = {
            "foo": self.foo,
            "predictors": self.predictors,
            "predictors_op": self.predictors_op,
            "dependent": self.dependent,
            "time_variable": self.time_variable,
            "treatment_identifier": self.treatment_identifier,
            "controls_identifier": self.controls_identifier,
            "time_predictors_prior": self.time_predictors_prior,
            "time_optimize_ssr": self.time_optimize_ssr,
            "special_predictors": self.special_predictors,
        }

        self.assertRaises(ValueError, Dataprep, unit_variable="badval", **kwargs)

    def test_init_arg_time_variable(self):
        kwargs = {
            "foo": self.foo,
            "predictors": self.predictors,
            "predictors_op": self.predictors_op,
            "dependent": self.dependent,
            "unit_variable": self.unit_variable,
            "treatment_identifier": self.treatment_identifier,
            "controls_identifier": self.controls_identifier,
            "time_predictors_prior": self.time_predictors_prior,
            "time_optimize_ssr": self.time_optimize_ssr,
            "special_predictors": self.special_predictors,
        }

        self.assertRaises(ValueError, Dataprep, time_variable="badval", **kwargs)

    def test_init_multiple_rows(self):
        kwargs = {
            "predictors": self.predictors,
            "predictors_op": self.predictors_op,
            "dependent": self.dependent,
            "unit_variable": self.unit_variable,
            "time_variable": self.time_variable,
            "treatment_identifier": self.treatment_identifier,
            "controls_identifier": self.controls_identifier,
            "time_predictors_prior": self.time_predictors_prior,
            "time_optimize_ssr": self.time_optimize_ssr,
            "special_predictors": self.special_predictors,
        }

        self.assertRaises(
            ValueError,
            Dataprep,
            foo=pd.concat([self.foo, self.foo.iloc[0:1]], axis=0),
            **kwargs,
        )

    def test_init_arg_treatment_identifier(self):
        kwargs = {
            "foo": self.foo,
            "predictors": self.predictors,
            "predictors_op": self.predictors_op,
            "dependent": self.dependent,
            "unit_variable": self.unit_variable,
            "time_variable": self.time_variable,
            "controls_identifier": self.controls_identifier,
            "time_predictors_prior": self.time_predictors_prior,
            "time_optimize_ssr": self.time_optimize_ssr,
            "special_predictors": self.special_predictors,
        }

        self.assertRaises(ValueError, Dataprep, treatment_identifier="badval", **kwargs)
        self.assertRaises(
            ValueError, Dataprep, treatment_identifier=["badval"], **kwargs
        )

    def test_bad_time_periods_time_predictors_prior(self):
        kwargs = {
            "foo": self.foo,
            "predictors": self.predictors,
            "predictors_op": self.predictors_op,
            "dependent": self.dependent,
            "unit_variable": self.unit_variable,
            "time_variable": self.time_variable,
            "controls_identifier": self.controls_identifier,
            "time_optimize_ssr": self.time_optimize_ssr,
            "treatment_identifier": self.treatment_identifier,
            "special_predictors": self.special_predictors,
        }

        self.assertRaises(
            ValueError, Dataprep, time_predictors_prior=["2", "3"], **kwargs
        )

    def test_bad_time_periods_time_optimize_ssr(self):
        kwargs = {
            "foo": self.foo,
            "predictors": self.predictors,
            "predictors_op": self.predictors_op,
            "dependent": self.dependent,
            "unit_variable": self.unit_variable,
            "time_variable": self.time_variable,
            "controls_identifier": self.controls_identifier,
            "time_predictors_prior": self.time_predictors_prior,
            "treatment_identifier": self.treatment_identifier,
            "special_predictors": self.special_predictors,
        }

        self.assertRaises(ValueError, Dataprep, time_optimize_ssr=["2", "3"], **kwargs)

    def test_bad_time_periods_special_predictors(self):
        kwargs = {
            "foo": self.foo,
            "predictors": self.predictors,
            "predictors_op": self.predictors_op,
            "dependent": self.dependent,
            "unit_variable": self.unit_variable,
            "time_variable": self.time_variable,
            "controls_identifier": self.controls_identifier,
            "time_optimize_ssr": self.time_optimize_ssr,
            "time_predictors_prior": self.time_predictors_prior,
            "treatment_identifier": self.treatment_identifier,
        }

        self.assertRaises(
            ValueError,
            Dataprep,
            special_predictors=[("predictor1", ["2"], "mean")],
            **kwargs,
        )

    def test_init_arg_controls_identifier(self):
        kwargs = {
            "foo": self.foo,
            "predictors": self.predictors,
            "predictors_op": self.predictors_op,
            "dependent": self.dependent,
            "unit_variable": self.unit_variable,
            "time_variable": self.time_variable,
            "time_predictors_prior": self.time_predictors_prior,
            "time_optimize_ssr": self.time_optimize_ssr,
            "special_predictors": self.special_predictors,
        }

        self.assertRaises(
            TypeError,
            Dataprep,
            treatment_identifier=self.treatment_identifier,
            controls_identifier=1,
            **kwargs,
        )
        self.assertRaises(
            ValueError,
            Dataprep,
            treatment_identifier=self.treatment_identifier,
            controls_identifier=[1],
            **kwargs,
        )
        self.assertRaises(
            ValueError,
            Dataprep,
            treatment_identifier=self.treatment_identifier,
            controls_identifier=[5],
            **kwargs,
        )
        self.assertRaises(
            TypeError,
            Dataprep,
            treatment_identifier=self.treatment_identifier_list,
            controls_identifier=1,
            **kwargs,
        )
        self.assertRaises(
            ValueError,
            Dataprep,
            treatment_identifier=self.treatment_identifier_list,
            controls_identifier=[1],
            **kwargs,
        )
        self.assertRaises(
            ValueError,
            Dataprep,
            treatment_identifier=self.treatment_identifier_list,
            controls_identifier=[5],
            **kwargs,
        )

    def test_init_arg_special_predictors(self):
        kwargs = {
            "foo": self.foo,
            "predictors": self.predictors,
            "predictors_op": self.predictors_op,
            "dependent": self.dependent,
            "unit_variable": self.unit_variable,
            "time_variable": self.time_variable,
            "treatment_identifier": self.treatment_identifier,
            "controls_identifier": self.controls_identifier,
            "time_predictors_prior": self.time_predictors_prior,
            "time_optimize_ssr": self.time_optimize_ssr,
        }

        self.assertRaises(
            ValueError, Dataprep, special_predictors=[("predictor1", [1])], **kwargs
        )
        self.assertRaises(
            ValueError,
            Dataprep,
            special_predictors=[("badval", [1], "mean")],
            **kwargs,
        )
        self.assertRaises(
            ValueError,
            Dataprep,
            special_predictors=[("predictor1", [1], "badval")],
            **kwargs,
        )

    def test_make_covariate_mats(self):
        kwargs = {
            "foo": self.foo,
            "predictors": self.predictors,
            "dependent": self.dependent,
            "unit_variable": self.unit_variable,
            "time_variable": self.time_variable,
            "treatment_identifier": self.treatment_identifier,
            "controls_identifier": self.controls_identifier,
            "time_predictors_prior": self.time_predictors_prior,
            "time_optimize_ssr": self.time_optimize_ssr,
            "special_predictors": self.special_predictors,
        }

        # Non-special
        for op in AGG_OP:
            dataprep = Dataprep(predictors_op=op, **kwargs)
            X0, X1 = dataprep.make_covariate_mats()

            # Treated
            for predictor in self.predictors:
                mask_treated = (
                    self.foo[self.unit_variable] == self.treatment_identifier
                ) & (self.foo[self.time_variable].isin(self.time_predictors_prior))
                self.assertAlmostEqual(
                    self.foo[mask_treated][predictor].agg(op), X1.loc[predictor]
                )

            # Control
            for control in self.controls_identifier:
                mask_control = self.foo[self.unit_variable] == control
                for predictor in self.predictors:
                    mask = mask_control & self.foo[self.time_variable].isin(
                        self.time_predictors_prior
                    )
                    self.assertAlmostEqual(
                        self.foo[mask][predictor].agg(op), X0.loc[predictor, control]
                    )

        # Special
        dataprep = Dataprep(predictors_op="mean", **kwargs)
        X0, X1 = dataprep.make_covariate_mats()

        # Treated
        for idx, (predictor, time_period, op) in enumerate(self.special_predictors, 1):
            mask_treated = (
                self.foo[self.unit_variable] == self.treatment_identifier
            ) & (self.foo[self.time_variable].isin(time_period))
            column_name = f"special.{idx}.{predictor}"
            self.assertAlmostEqual(
                self.foo[mask_treated][predictor].agg(op), X1.loc[column_name]
            )

        # Control
        for control in self.controls_identifier:
            mask_control = self.foo[self.unit_variable] == control
            for idx, (predictor, time_period, op) in enumerate(
                self.special_predictors, 1
            ):
                mask = mask_control & self.foo[self.time_variable].isin(time_period)
                column_name = f"special.{idx}.{predictor}"
                self.assertAlmostEqual(
                    self.foo[mask][predictor].agg(op), X0.loc[column_name, control]
                )

    def test_make_covariate_mats_list_single(self):
        kwargs = {
            "foo": self.foo,
            "predictors": self.predictors,
            "predictors_op": self.predictors_op,
            "dependent": self.dependent,
            "unit_variable": self.unit_variable,
            "time_variable": self.time_variable,
            "time_predictors_prior": self.time_predictors_prior,
            "time_optimize_ssr": self.time_optimize_ssr,
            "special_predictors": self.special_predictors,
        }
        dataprep = Dataprep(
            treatment_identifier=self.treatment_identifier,
            controls_identifier=self.controls_identifier,
            **kwargs,
        )
        X0, X1 = dataprep.make_covariate_mats()
        self.assertIsInstance(X0, pd.DataFrame)
        self.assertIsInstance(X1, pd.Series)

        dataprep = Dataprep(
            treatment_identifier=[self.treatment_identifier],
            controls_identifier=self.controls_identifier,
            **kwargs,
        )
        X0, X1 = dataprep.make_covariate_mats()
        self.assertIsInstance(X0, pd.DataFrame)
        self.assertIsInstance(X1, pd.Series)

        dataprep = Dataprep(
            treatment_identifier=self.treatment_identifier_list,
            controls_identifier=self.controls_identifier_alt,
            **kwargs,
        )
        X0, X1 = dataprep.make_covariate_mats()
        self.assertIsInstance(X0, pd.DataFrame)
        self.assertIsInstance(X1, pd.DataFrame)

    def test_make_outcome_mats(self):
        kwargs = {
            "foo": self.foo,
            "predictors": self.predictors,
            "predictors_op": self.predictors_op,
            "dependent": self.dependent,
            "unit_variable": self.unit_variable,
            "time_variable": self.time_variable,
            "treatment_identifier": self.treatment_identifier,
            "controls_identifier": self.controls_identifier,
            "time_predictors_prior": self.time_predictors_prior,
            "time_optimize_ssr": self.time_optimize_ssr,
            "special_predictors": self.special_predictors,
        }

        dataprep = Dataprep(**kwargs)
        # All time
        all_time = dataprep.foo[dataprep.time_variable].unique()
        Z0, Z1 = dataprep.make_outcome_mats(time_period=all_time)

        # Treated
        mask_treated = self.foo[self.unit_variable] == self.treatment_identifier
        pd.testing.assert_series_equal(
            self.foo[mask_treated]
            .set_index(self.time_variable)[self.dependent]
            .rename(self.treatment_identifier),
            Z1,
        )

        # Control
        for control in self.controls_identifier:
            mask = self.foo[self.unit_variable] == control
            pd.testing.assert_series_equal(
                self.foo[mask]
                .set_index(self.time_variable)[self.dependent]
                .rename(control),
                Z0[control],
            )

        # Set time
        Z0, Z1 = dataprep.make_outcome_mats()

        # Treated
        mask_treated = (self.foo[self.unit_variable] == self.treatment_identifier) & (
            self.foo[self.time_variable].isin(self.time_optimize_ssr)
        )
        pd.testing.assert_series_equal(
            self.foo[mask_treated]
            .set_index(self.time_variable)[self.dependent]
            .rename(self.treatment_identifier),
            Z1,
        )

        # Control
        for control in self.controls_identifier:
            mask = (self.foo[self.unit_variable] == control) & (
                self.foo[self.time_variable].isin(self.time_optimize_ssr)
            )
            pd.testing.assert_series_equal(
                self.foo[mask]
                .set_index(self.time_variable)[self.dependent]
                .rename(control),
                Z0[control],
            )

    def test_make_outcome_mats_list_single(self):
        kwargs = {
            "foo": self.foo,
            "predictors": self.predictors,
            "predictors_op": self.predictors_op,
            "dependent": self.dependent,
            "unit_variable": self.unit_variable,
            "time_variable": self.time_variable,
            "time_predictors_prior": self.time_predictors_prior,
            "time_optimize_ssr": self.time_optimize_ssr,
            "special_predictors": self.special_predictors,
        }
        dataprep = Dataprep(
            treatment_identifier=self.treatment_identifier,
            controls_identifier=self.controls_identifier,
            **kwargs,
        )
        Z0, Z1 = dataprep.make_outcome_mats()
        self.assertIsInstance(Z0, pd.DataFrame)
        self.assertIsInstance(Z1, pd.Series)

        dataprep = Dataprep(
            treatment_identifier=[self.treatment_identifier],
            controls_identifier=self.controls_identifier,
            **kwargs,
        )
        Z0, Z1 = dataprep.make_outcome_mats()
        self.assertIsInstance(Z0, pd.DataFrame)
        self.assertIsInstance(Z1, pd.Series)

        dataprep = Dataprep(
            treatment_identifier=self.treatment_identifier_list,
            controls_identifier=self.controls_identifier_alt,
            **kwargs,
        )
        Z0, Z1 = dataprep.make_outcome_mats()
        self.assertIsInstance(Z0, pd.DataFrame)
        self.assertIsInstance(Z1, pd.DataFrame)

    def test_str(self):
        kwargs_no_special = {
            "foo": self.foo,
            "predictors": self.predictors,
            "predictors_op": self.predictors_op,
            "dependent": self.dependent,
            "unit_variable": self.unit_variable,
            "time_variable": self.time_variable,
            "treatment_identifier": self.treatment_identifier,
            "controls_identifier": self.controls_identifier,
            "time_predictors_prior": self.time_predictors_prior,
            "time_optimize_ssr": self.time_optimize_ssr,
        }
        dataprep = Dataprep(**kwargs_no_special)
        self.assertEqual(self.str_no_special, str(dataprep))

        kwargs_special = {
            "foo": self.foo,
            "predictors": self.predictors,
            "predictors_op": self.predictors_op,
            "dependent": self.dependent,
            "unit_variable": self.unit_variable,
            "time_variable": self.time_variable,
            "treatment_identifier": self.treatment_identifier,
            "controls_identifier": self.controls_identifier,
            "time_predictors_prior": self.time_predictors_prior,
            "time_optimize_ssr": self.time_optimize_ssr,
            "special_predictors": self.special_predictors,
        }
        dataprep = Dataprep(**kwargs_special)
        self.assertEqual(self.str_special, str(dataprep))



import unittest
from unittest.mock import patch, Mock
import numpy as np
import pandas as pd

import pysyncon


class TestSynth(unittest.TestCase):
    def setUp(self):
        self.foo = pd.DataFrame(
            {
                "time": [1, 2, 3, 4, 1, 2, 3, 4, 1, 2, 3, 4],
                "name": [1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3],
                "dependent": np.random.random(12),
                "predictor1": np.random.random(12),
                "predictor2": np.random.random(12),
            }
        )
        self.predictors = ["predictor1"]
        self.predictors_op = "mean"
        self.dependent = "dependent"
        self.unit_variable = "name"
        self.time_variable = "time"
        self.treatment_identifier = 1
        self.treatment_identifier_list = [1, 2]
        self.controls_identifier = [2, 3]
        self.controls_identifier_alt = [3]
        self.time_predictors_prior = [2, 3]
        self.time_optimize_ssr = [1, 2, 3]
        self.special_predictors = [
            ("predictor1", [2], "mean"),
            ("predictor2", [1, 2], "median"),
            ("predictor2", [1, 2], "std"),
        ]

    def test_fit_treated(self):
        kwargs = {
            "foo": self.foo,
            "predictors": self.predictors,
            "predictors_op": self.predictors_op,
            "dependent": self.dependent,
            "unit_variable": self.unit_variable,
            "time_variable": self.time_variable,
            "time_predictors_prior": self.time_predictors_prior,
            "time_optimize_ssr": self.time_optimize_ssr,
            "special_predictors": self.special_predictors,
        }
        dataprep = pysyncon.Dataprep(
            treatment_identifier=self.treatment_identifier_list,
            controls_identifier=self.controls_identifier_alt,
            **kwargs,
        )
        synth = pysyncon.Synth()
        self.assertRaises(ValueError, synth.fit, dataprep)

        dataprep = pysyncon.Dataprep(
            treatment_identifier=self.treatment_identifier,
            controls_identifier=self.controls_identifier,
            **kwargs,
        )
        synth = pysyncon.Synth()

        # Run with normal controls list
        synth.fit(dataprep)

        dataprep = pysyncon.Dataprep(
            treatment_identifier=[self.treatment_identifier],
            controls_identifier=self.controls_identifier,
            **kwargs,
        )
        synth = pysyncon.Synth()

        # Run with a list of treatment identifiers
        synth.fit(dataprep)

    def test_X0_X1_fit(self):
        synth = pysyncon.Synth()

        # Neither dataprep nor matrices set
        self.assertRaises(ValueError, synth.fit)

        # X1 needs to be pd.Series
        X0 = pd.DataFrame(np.random.rand(5, 5))
        X1 = pd.DataFrame(np.random.rand(5, 2))
        Z0 = pd.DataFrame(np.random.rand(5, 5))
        Z1 = pd.DataFrame(np.random.rand(5, 2))
        self.assertRaises(TypeError, synth.fit, X0=X0, X1=X1, Z0=Z0, Z1=Z1)

        # X1 needs to be pd.Series
        X0 = pd.DataFrame(np.random.rand(5, 5))
        X1 = pd.DataFrame(np.random.rand(5, 1))
        Z0 = pd.DataFrame(np.random.rand(5, 5))
        Z1 = pd.DataFrame(np.random.rand(5, 1))
        self.assertRaises(TypeError, synth.fit, X0=X0, X1=X1, Z0=Z0, Z1=Z1)

    @patch("pysyncon.base.plt")
    def test_path_plot(self, mock_plt: Mock):
        kwargs = {
            "foo": self.foo,
            "predictors": self.predictors,
            "predictors_op": self.predictors_op,
            "dependent": self.dependent,
            "unit_variable": self.unit_variable,
            "time_variable": self.time_variable,
            "treatment_identifier": self.treatment_identifier,
            "controls_identifier": self.controls_identifier,
            "time_predictors_prior": self.time_predictors_prior,
            "time_optimize_ssr": self.time_optimize_ssr,
            "special_predictors": self.special_predictors,
        }

        dataprep = pysyncon.Dataprep(**kwargs)
        synth = pysyncon.Synth()
        # No weight matrix set
        self.assertRaises(ValueError, synth.path_plot)

        X0, X1 = dataprep.make_covariate_mats()
        Z0, Z1 = dataprep.make_outcome_mats()
        synth.fit(X0=X0, X1=X1, Z0=Z0, Z1=Z1)
        # No Dataprep object available
        self.assertRaises(ValueError, synth.path_plot)

        synth.fit(dataprep=dataprep)
        synth.path_plot()

        self.assertEqual(mock_plt.plot.call_count, 2)
        first_call, second_call = mock_plt.plot.call_args_list

        _, first_call_kwargs = first_call
        self.assertEqual(first_call_kwargs["color"], "black")
        self.assertEqual(first_call_kwargs["linewidth"], 1)
        self.assertEqual(first_call_kwargs["label"], dataprep.treatment_identifier)

        _, second_call_kwargs = second_call
        self.assertEqual(second_call_kwargs["color"], "black")
        self.assertEqual(second_call_kwargs["linewidth"], 1)
        self.assertEqual(second_call_kwargs["linestyle"], "dashed")
        self.assertEqual(second_call_kwargs["label"], "Synthetic")

        mock_plt.axvline.assert_not_called()
        mock_plt.legend.assert_called()
        mock_plt.grid.assert_called_with(True)
        mock_plt.show.assert_called()

        synth.path_plot(treatment_time=3)
        mock_plt.axvline.assert_called_once()

        _, kwargs = mock_plt.axvline.call_args
        self.assertEqual(kwargs["x"], 3)
        self.assertEqual(kwargs["ymin"], 0.05)
        self.assertEqual(kwargs["ymax"], 0.95)
        self.assertEqual(kwargs["linestyle"], "dashed")

    @patch("pysyncon.base.plt")
    def test_gaps_plot(self, mock_plt: Mock):
        kwargs = {
            "foo": self.foo,
            "predictors": self.predictors,
            "predictors_op": self.predictors_op,
            "dependent": self.dependent,
            "unit_variable": self.unit_variable,
            "time_variable": self.time_variable,
            "treatment_identifier": self.treatment_identifier,
            "controls_identifier": self.controls_identifier,
            "time_predictors_prior": self.time_predictors_prior,
            "time_optimize_ssr": self.time_optimize_ssr,
            "special_predictors": self.special_predictors,
        }

        dataprep = pysyncon.Dataprep(**kwargs)
        synth = pysyncon.Synth()
        # No weight matrix set
        self.assertRaises(ValueError, synth.gaps_plot)

        X0, X1 = dataprep.make_covariate_mats()
        Z0, Z1 = dataprep.make_outcome_mats()
        synth.fit(X0=X0, X1=X1, Z0=Z0, Z1=Z1)
        # No Dataprep object available
        self.assertRaises(ValueError, synth.gaps_plot)

        synth.fit(dataprep=dataprep)
        synth.gaps_plot()

        self.assertEqual(mock_plt.plot.call_count, 1)
        _, kwargs = mock_plt.plot.call_args

        self.assertEqual(kwargs["color"], "black")
        self.assertEqual(kwargs["linewidth"], 1)

        mock_plt.axvline.assert_not_called()
        mock_plt.grid.assert_called_with(True)
        mock_plt.show.assert_called()

        synth.path_plot(treatment_time=3)
        mock_plt.axvline.assert_called_once()

        _, kwargs = mock_plt.axvline.call_args
        self.assertEqual(kwargs["x"], 3)
        self.assertEqual(kwargs["ymin"], 0.05)
        self.assertEqual(kwargs["ymax"], 0.95)
        self.assertEqual(kwargs["linestyle"], "dashed")

    def test_weight(self):
        synth = pysyncon.Synth()
        # No weight matrix set
        self.assertRaises(ValueError, synth.weights)

    def test_summary(self):
        kwargs = {
            "foo": self.foo,
            "predictors": self.predictors,
            "predictors_op": self.predictors_op,
            "dependent": self.dependent,
            "unit_variable": self.unit_variable,
            "time_variable": self.time_variable,
            "treatment_identifier": self.treatment_identifier,
            "controls_identifier": self.controls_identifier,
            "time_predictors_prior": self.time_predictors_prior,
            "time_optimize_ssr": self.time_optimize_ssr,
            "special_predictors": self.special_predictors,
        }

        dataprep = pysyncon.Dataprep(**kwargs)
        synth = pysyncon.Synth()
        # No weight matrix set
        self.assertRaises(ValueError, synth.summary)
        X0, X1 = dataprep.make_covariate_mats()
        Z0, Z1 = dataprep.make_outcome_mats()
        synth.fit(X0=X0, X1=X1, Z0=Z0, Z1=Z1)
        # No Dataprep object available
        self.assertRaises(ValueError, synth.summary)

        synth.V = None
        # No V matrix available
        self.assertRaises(ValueError, synth.summary)

    def test_att(self):
        synth = pysyncon.Synth()
        # No weight matrix set
        self.assertRaises(ValueError, synth.att, range(1))

    def test_metrics(self):
        kwargs = {
            "foo": self.foo,
            "predictors": self.predictors,
            "predictors_op": self.predictors_op,
            "dependent": self.dependent,
            "unit_variable": self.unit_variable,
            "time_variable": self.time_variable,
            "treatment_identifier": self.treatment_identifier,
            "controls_identifier": self.controls_identifier,
            "time_predictors_prior": self.time_predictors_prior,
            "time_optimize_ssr": self.time_optimize_ssr,
            "special_predictors": self.special_predictors,
        }

        dataprep = pysyncon.Dataprep(**kwargs)
        synth = pysyncon.Synth()

        X0, X1 = dataprep.make_covariate_mats()
        Z0, Z1 = dataprep.make_outcome_mats()
        synth.fit(X0=X0, X1=X1, Z0=Z0, Z1=Z1)
        # No Dataprep object available
        self.assertRaises(ValueError, synth.mspe)
        self.assertRaises(ValueError, synth.mape)
        self.assertRaises(ValueError, synth.mae)

        del synth

        synth = pysyncon.Synth()
        synth.dataprep = dataprep
        # No weights availble/fit not run available
        self.assertRaises(ValueError, synth.mspe)
        self.assertRaises(ValueError, synth.mape)
        self.assertRaises(ValueError, synth.mae)

    def test_confidence_intervals(self):
        kwargs = {
            "foo": self.foo,
            "predictors": self.predictors,
            "predictors_op": self.predictors_op,
            "dependent": self.dependent,
            "unit_variable": self.unit_variable,
            "time_variable": self.time_variable,
            "treatment_identifier": self.treatment_identifier,
            "controls_identifier": self.controls_identifier,
            "time_predictors_prior": self.time_predictors_prior,
            "time_optimize_ssr": self.time_optimize_ssr,
            "special_predictors": self.special_predictors,
        }

        dataprep = pysyncon.Dataprep(**kwargs)
        synth = pysyncon.Synth()
        synth.fit(dataprep=dataprep)

        # Bad option
        self.assertRaises(
            ValueError,
            synth.confidence_interval,
            alpha=0.5,
            time_periods=[4],
            tol=0.01,
            method="foo",
        )

        # Run with dataprep supplied
        synth.confidence_interval(
            alpha=0.5, time_periods=[4], dataprep=dataprep, tol=0.01
        )

        # Too few time periods for alpha value
        self.assertRaises(
            ValueError,
            synth.confidence_interval,
            alpha=0.05,
            time_periods=[4],
            tol=0.01,
            dataprep=dataprep,
        )

        # Run without dataprep supplied
        synth.confidence_interval(alpha=0.5, time_periods=[4], tol=0.01)

        # Too few time periods for alpha value
        self.assertRaises(
            ValueError,
            synth.confidence_interval,
            alpha=0.05,
            time_periods=[4],
            tol=0.01,
        )

        # Without dataprep supplied or matrices
        synth.dataprep = None
        self.assertRaises(
            ValueError, synth.confidence_interval, alpha=0.5, time_periods=[4], tol=0.01
        )

        # No pre-periods supplied
        synth.dataprep = None
        X0, X1 = dataprep.make_covariate_mats()
        Z0, Z1 = dataprep.make_outcome_mats(time_period=[1, 2, 3, 4])
        self.assertRaises(
            ValueError,
            synth.confidence_interval,
            alpha=0.5,
            time_periods=[4],
            tol=0.01,
            X0=X0,
            X1=X1,
            Z0=Z0,
            Z1=Z1,
        )

        # Bad alpha value
        self.assertRaises(
            ValueError,
            synth.confidence_interval,
            alpha=0.05,
            time_periods=[4],
            pre_periods=[1, 2, 3],
            tol=0.01,
            X0=X0,
            X1=X1,
            Z0=Z0,
            Z1=Z1,
        )

        # Dataframes supplied instead of series
        X1 = X1.to_frame()
        Z1 = Z1.to_frame()
        self.assertRaises(
            TypeError,
            synth.confidence_interval,
            alpha=0.5,
            time_periods=[4],
            pre_periods=[1, 2, 3],
            tol=0.01,
            X0=X0,
            X1=X1,
            Z0=Z0,
            Z1=Z1,
        )



import unittest
import pandas as pd
import numpy as np

from pysyncon import Synth
from pysyncon.inference import ConformalInference


class TestConformalInference(unittest.TestCase):
    def setUp(self):
        self.rng = np.random.default_rng()
        self.alpha = 0.05
        self.scm = Synth()
        self.Z0 = pd.DataFrame(
            data=self.rng.random(size=(30, 10)),
            index=range(1, 31),
            columns=range(1, 11),
        )
        self.Z1 = pd.Series(
            data=self.rng.random(size=(30,)), index=range(1, 31), name=0
        )
        self.X0 = pd.DataFrame(
            data=self.rng.random(size=(4, 10)),
            index=range(1, 5),
            columns=range(1, 11),
        )
        self.X1 = pd.Series(data=self.rng.random(size=(4,)), index=range(1, 5), name=0)
        self.pre_periods = list(range(1, 21))
        self.post_periods = list(range(21, 31))
        self.max_iter = 20
        self.tol = 0.1
        self.step_sz = None
        self.step_sz_div = 20.0
        self.verbose = False

    def test_alpha(self):
        kwargs = {
            "scm": self.scm,
            "Z0": self.Z0,
            "Z1": self.Z1,
            "pre_periods": self.pre_periods,
            "post_periods": self.post_periods,
            "max_iter": self.max_iter,
            "tol": self.tol,
            "step_sz": self.step_sz,
            "step_sz_div": self.step_sz_div,
            "verbose": self.verbose,
        }

        conformal_inf = ConformalInference()

        cases = [-1.0, 0.0, 1.0, 17.0]
        for case in cases:
            with self.subTest(case=case):
                self.assertRaises(
                    ValueError, conformal_inf.confidence_intervals, alpha=case, **kwargs
                )

        cases = [True, ["foo"], {"foo": "bar"}]
        for case in cases:
            with self.subTest(case=case):
                self.assertRaises(
                    TypeError, conformal_inf.confidence_intervals, alpha=case, **kwargs
                )

    def test_max_iter(self):
        kwargs = {
            "alpha": self.alpha,
            "scm": self.scm,
            "Z0": self.Z0,
            "Z1": self.Z1,
            "pre_periods": self.pre_periods,
            "post_periods": self.post_periods,
            "tol": self.tol,
            "step_sz": self.step_sz,
            "step_sz_div": self.step_sz_div,
            "verbose": self.verbose,
        }

        conformal_inf = ConformalInference()

        cases = [-17, 0]
        for case in cases:
            with self.subTest(case=case):
                self.assertRaises(
                    ValueError,
                    conformal_inf.confidence_intervals,
                    max_iter=case,
                    **kwargs
                )

        cases = [5.2, 10.0]
        for case in cases:
            with self.subTest(case=case):
                self.assertRaises(
                    TypeError,
                    conformal_inf.confidence_intervals,
                    max_iter=case,
                    **kwargs
                )

    def test_tol(self):
        kwargs = {
            "alpha": self.alpha,
            "scm": self.scm,
            "Z0": self.Z0,
            "Z1": self.Z1,
            "pre_periods": self.pre_periods,
            "post_periods": self.post_periods,
            "max_iter": self.max_iter,
            "step_sz": self.step_sz,
            "step_sz_div": self.step_sz_div,
            "verbose": self.verbose,
        }

        conformal_inf = ConformalInference()

        cases = [-4.2, 0.0]
        for case in cases:
            with self.subTest(case=case):
                self.assertRaises(
                    ValueError, conformal_inf.confidence_intervals, tol=case, **kwargs
                )

        cases = [-4, 0]
        for case in cases:
            with self.subTest(case=case):
                self.assertRaises(
                    TypeError, conformal_inf.confidence_intervals, tol=case, **kwargs
                )

    def test_step_sz(self):
        kwargs = {
            "alpha": self.alpha,
            "scm": self.scm,
            "Z0": self.Z0,
            "Z1": self.Z1,
            "pre_periods": self.pre_periods,
            "post_periods": self.post_periods,
            "tol": self.tol,
            "max_iter": self.max_iter,
            "step_sz_div": self.step_sz_div,
            "verbose": self.verbose,
        }

        conformal_inf = ConformalInference()

        cases = [-4.2, 0.0]
        for case in cases:
            with self.subTest(case=case):
                self.assertRaises(
                    ValueError,
                    conformal_inf.confidence_intervals,
                    step_sz=case,
                    **kwargs
                )

        cases = [-4, 0]
        for case in cases:
            with self.subTest(case=case):
                self.assertRaises(
                    TypeError,
                    conformal_inf.confidence_intervals,
                    step_sz=case,
                    **kwargs
                )

    def test_step_sz_tol(self):
        kwargs = {
            "alpha": self.alpha,
            "scm": self.scm,
            "Z0": self.Z0,
            "Z1": self.Z1,
            "pre_periods": self.pre_periods,
            "post_periods": self.post_periods,
            "max_iter": self.max_iter,
            "step_sz_div": self.step_sz_div,
            "verbose": self.verbose,
        }

        conformal_inf = ConformalInference()

        # Step-size is less than tolerance
        self.assertRaises(
            ValueError,
            conformal_inf.confidence_intervals,
            tol=1.0,
            step_sz=0.1,
            **kwargs
        )

        # Step-size = tolerance
        self.assertRaises(
            ValueError,
            conformal_inf.confidence_intervals,
            tol=1.0,
            step_sz=1.0,
            **kwargs
        )

    def test_step_sz_guessing(self):
        kwargs = {
            "alpha": self.alpha,
            "scm": self.scm,
            "Z0": self.Z0,
            "Z1": self.Z1,
            "pre_periods": self.pre_periods,
            "post_periods": self.post_periods,
            "max_iter": self.max_iter,
            "step_sz_div": self.step_sz_div,
            "verbose": self.verbose,
            "scm_fit_args": {"X0": self.X0, "X1": self.X1},
        }

        conformal_inf = ConformalInference()

        # No step-size and a big tolerance
        # (step-size guessing)
        _, n_c = self.Z0.shape
        self.scm.W = np.full(n_c, 1.0 / n_c)
        conformal_inf.confidence_intervals(tol=100.0, **kwargs)
        self.scm.W = None

    def test_step_sz_div(self):
        kwargs = {
            "alpha": self.alpha,
            "scm": self.scm,
            "Z0": self.Z0,
            "Z1": self.Z1,
            "pre_periods": self.pre_periods,
            "post_periods": self.post_periods,
            "tol": self.tol,
            "max_iter": self.max_iter,
            "step_sz": self.step_sz,
            "verbose": self.verbose,
        }

        conformal_inf = ConformalInference()

        cases = [-4.2, 0.0]
        for case in cases:
            with self.subTest(case=case):
                self.assertRaises(
                    ValueError,
                    conformal_inf.confidence_intervals,
                    step_sz_div=case,
                    **kwargs
                )

        cases = [-4, 0]
        for case in cases:
            with self.subTest(case=case):
                self.assertRaises(
                    TypeError,
                    conformal_inf.confidence_intervals,
                    step_sz_div=case,
                    **kwargs
                )

    def test_no_weights(self):
        kwargs = {
            "alpha": self.alpha,
            "scm": self.scm,
            "Z0": self.Z0,
            "Z1": self.Z1,
            "pre_periods": self.pre_periods,
            "post_periods": self.post_periods,
            "tol": self.tol,
            "max_iter": self.max_iter,
            "step_sz": self.step_sz,
            "verbose": self.verbose,
        }

        conformal_inf = ConformalInference()
        self.assertRaises(ValueError, conformal_inf.confidence_intervals, **kwargs)

    def test_root_search(self):
        cases_roots_x0 = [
            ((-1, 3), 0.5),
            ((-1, 3), 1.0),
            ((-1, 3), 2.5),
            ((-1, 400), 0.5),
            ((-1, 400), 100),
            ((-1, 400), 399),
        ]
        cases_step_sz = [0.1, 1.0]

        ci = ConformalInference()
        tol = 0.01
        for case_root_x0 in cases_roots_x0:
            for case_step_sz in cases_step_sz:
                case = (case_root_x0, case_step_sz)
                with self.subTest(case=case):
                    ((lower, upper), x0) = case_root_x0

                    res = ci._root_search(
                        fn=lambda x: (lower - x) * (x - upper),
                        x0=x0,
                        direction=-1,
                        tol=tol,
                        step_sz=case_step_sz,
                        max_iter=100,
                    )
                    self.assertAlmostEqual(lower, res, delta=tol)

                    res = ci._root_search(
                        fn=lambda x: (lower - x) * (x - upper),
                        x0=x0,
                        direction=1,
                        tol=tol,
                        step_sz=case_step_sz,
                        max_iter=100,
                    )
                    self.assertAlmostEqual(upper, res, delta=tol)

        self.assertRaises(
            Exception,
            ci._root_search,
            fn=lambda x: (-1 - x) * (x - 400),
            x0=200,
            direction=-1,
            tol=0.01,
            step_sz=1.0,
            max_iter=1,
        )



from __future__ import annotations
from typing import Optional, Union
from concurrent import futures
import copy
from dataclasses import dataclass

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from .dataprep import Dataprep, IsinArg_t
from .base import BaseSynth


class HoldoutSplitter:
    """Iterator that prepares the time series for cross-validation by
    progressively removing blocks of length `holdout_len`.
    """

    def __init__(self, df: pd.DataFrame, ser: pd.Series, holdout_len: int = 1):
        """Iterator that prepares the time series for cross-validation by
        progressively removing blocks of length `holdout_len`.

        Parameters
        ----------
        df : pandas.DataFrame, shape (r, c)
            Dataframe that will be split for the cross-validation.
        ser : pandas.Series, shape (r, 1)
            Series that will split for the cross-validation.
        holdout_len : int, optional
            Number of days to remove in each iteration, by default 1.

        Raises
        ------
        ValueError
            if df and ser do not have the same number of rows.
        ValueError
            if `holdout_len` is not >= 1.
        ValueError
            if `holdout_len` is larger than the number of rows of df.
        """
        if df.shape[0] != ser.shape[0]:
            raise ValueError("`df` and `ser` must have the same number of rows.")
        if holdout_len < 1:
            raise ValueError("`holdout_len` must be at least 1.")
        if holdout_len >= df.shape[0]:
            raise ValueError("`holdout_len` must be less than df.shape[0]")
        self.df = df
        self.ser = ser
        self.holdout_len = holdout_len
        self.idx = 0

    def __iter__(self):
        self.idx = 0
        return self

    def __next__(self) -> tuple[pd.DataFrame, pd.DataFrame, pd.Series, pd.Series]:
        if (self.idx + self.holdout_len) > self.df.shape[0]:
            raise StopIteration
        holdout = slice(self.idx, self.idx + self.holdout_len)

        df_holdout = self.df.iloc[holdout,]  # fmt: skip
        ser_holdout = self.ser.iloc[holdout]

        df = self.df.drop(index=self.df.index[holdout])
        ser = self.ser.drop(index=self.ser.index[holdout])

        self.idx += 1
        return df, df_holdout, ser, ser_holdout


@dataclass
class CrossValidationResult:
    """Convenience class for holding the results of the cross-validation
    procedure from the AugSynth.
    """

    lambdas: np.ndarray
    errors_mean: np.ndarray
    errors_se: np.ndarray

    def best_lambda(self, min_1se: bool = True) -> float:
        """Return the best lambda.

        Parameters
        ----------
        min_1se : bool, optional
            return the largest lambda within 1 standard error of the minimum
            , by default True

        Returns
        -------
        float
        """
        if min_1se:
            return (
                self.lambdas[
                    self.errors_mean
                    <= self.errors_mean.min()
                    + self.errors_se[self.errors_mean.argmin()]
                ]
                .max()
                .item()
            )
        return self.lambdas[self.errors_mean.argmin()].item()

    def plot(self) -> None:
        """Plots the mean errors against the lambda values with the standard
        errors as error bars.
        """
        plt.errorbar(
            x=self.lambdas,
            y=self.errors_mean,
            yerr=self.errors_se,
            ecolor="black",
            capsize=2,
        )
        plt.xlabel("Lambda")
        plt.ylabel("Mean error")
        plt.xscale("log")
        plt.yscale("log")
        plt.title("Cross validation result")
        plt.grid()
        plt.show()


class PlaceboTest:
    """Class that carries out placebo tests by running a synthetic control
    study using each possible control unit as the treated unit and the
    remaining control units as controls. See :cite:`germany2015` for more details.
    """

    def __init__(self) -> None:
        self.paths: Optional[pd.DataFrame] = None
        self.treated_path: Optional[pd.DataFrame] = None
        self.gaps: Optional[pd.DataFrame] = None
        self.treated_gap: Optional[pd.DataFrame] = None
        self.time_optimize_ssr: Optional[IsinArg_t] = None

    def fit(
        self,
        dataprep: Dataprep,
        scm: BaseSynth,
        scm_options: dict = {},
        max_workers: Optional[int] = None,
        verbose: bool = True,
    ):
        """Run the placebo tests. This method is multi-process and by default
        will use all available processors. Use the `max_workers` option to change
        this behaviour.

        Parameters
        ----------
        dataprep : Dataprep
            :class:`Dataprep` object containing data to model, by default None.
        scm : Synth | AugSynth
            Synthetic control study to use
        scm_options : dict, optional
            Options to provide to the fit method of the synthetic control
            study, valid options are any valid option that the `scm_type`
            takes, by default {}
        max_workers : Optional[int], optional
            Maximum number of processes to use, if not provided then will use
            all available, by default None
        verbose : bool, optional
            Whether or not to output progress, by default True
        """
        paths, gaps = list(), list()
        n_tests = len(dataprep.controls_identifier)
        with futures.ProcessPoolExecutor(max_workers=max_workers) as executor:
            to_do = list()
            for treated, controls in self.placebo_iter(dataprep.controls_identifier):
                _dataprep = copy.copy(dataprep)
                _dataprep.treatment_identifier = treated
                _dataprep.controls_identifier = controls
                to_do.append(
                    executor.submit(
                        self._single_placebo,
                        dataprep=_dataprep,
                        scm=scm,
                        scm_options=scm_options,
                    )
                )
            for idx, future in enumerate(futures.as_completed(to_do), 1):
                path, gap = future.result()
                if verbose:
                    print(f"({idx}/{n_tests}) Completed placebo test for {path.name}.")
                paths.append(path)
                gaps.append(gap)

        self.paths = pd.concat(paths, axis=1)
        self.gaps = pd.concat(gaps, axis=1)
        self.time_optimize_ssr = dataprep.time_optimize_ssr

        print(f"Calculating treated unit gaps.")
        self.treated_path, self.treated_gap = self._single_placebo(
            dataprep=dataprep, scm=scm, scm_options=scm_options
        )
        print("Done.")

    @staticmethod
    def placebo_iter(controls: list[str]) -> tuple[str, list[str]]:
        """Generates combinations of (treated unit, control units) for the
        placebo tests.

        Parameters
        ----------
        controls : list[str]
            List of unit labels to use

        Yields
        ------
        tuple[str, list[str]]
            Tuple of (treated unit label, control unit labels)

        :meta private:
        """
        for control in controls:
            yield (control, [c for c in controls if c != control])

    @staticmethod
    def _single_placebo(
        dataprep: Dataprep, scm: BaseSynth, scm_options: dict = {}
    ) -> tuple[pd.Series, pd.Series]:
        """Run a single placebo test.

        Parameters
        ----------
        dataprep : Dataprep
            :class:`Dataprep` object containing data to model
        scm : Synth | AugSynth
            Type of synthetic control study to use
        scm_options : dict, optional
            Options to provide to the fit method of the synthetic control
            study, valid options are any valid option that `scm` takes, by
            default {}

        Returns
        -------
        tuple[pandas.Series, pandas.Series]
            A time-series of the path of the synthetic control and a
            time-series of the gap between the treated unit and the synthetic
            control.

        :meta private:
        """
        scm.fit(dataprep=dataprep, **scm_options)

        Z0, Z1 = dataprep.make_outcome_mats(
            time_period=dataprep.foo[dataprep.time_variable]
        )
        synthetic = scm._synthetic(Z0=Z0)
        gaps = scm._gaps(Z0=Z0, Z1=Z1)
        return synthetic.rename(dataprep.treatment_identifier), gaps.rename(
            dataprep.treatment_identifier
        )

    def gaps_plot(
        self,
        time_period: Optional[IsinArg_t] = None,
        grid: bool = True,
        treatment_time: Optional[int] = None,
        mspe_threshold: Optional[float] = None,
        exclude_units: Optional[list] = None,
    ):
        """Plot the gaps between the treated unit and the synthetic control
        for each placebo test.

        Parameters
        ----------
        time_period : Iterable | pandas.Series | dict, optional
            Time range to plot, if none is supplied then the time range used
            is the time period over which the optimisation happens, by default
            None
        grid : bool, optional
            Whether or not to plot a grid, by default True
        treatment_time : int, optional
            If supplied, plot a vertical line at the time period that the
            treatment time occurred, by default None
        mspe_threshold : float, optional
            Remove any non-treated units whose MSPE pre-treatment is :math:`>`
            mspe_threshold :math:`\\times` the MSPE of the treated unit pre-treatment.
            This serves to exclude any non-treated units whose synthetic control
            had a poor pre-treatment match to the actual relative to how the
            actual treated unit matched pre-treatment.

        Raises
        ------
        ValueError
            if no placebo test has been run yet
        ValueError
            if `mspe_threshold` is supplied but `treatment_year` is not.
        """
        if self.gaps is None:
            raise ValueError("No gaps available; run a placebo test first.")
        time_period = time_period if time_period is not None else self.time_optimize_ssr

        gaps = self.gaps.drop(columns=exclude_units) if exclude_units else self.gaps

        if mspe_threshold:
            if not treatment_time:
                raise ValueError("Need `treatment_time` to use `mspe_threshold`.")
            pre_mspe = gaps.loc[:treatment_time].pow(2).sum(axis=0)
            pre_mspe_treated = self.treated_gap.loc[:treatment_time].pow(2).sum(axis=0)
            keep = pre_mspe[pre_mspe < mspe_threshold * pre_mspe_treated].index
            placebo_gaps = gaps[gaps.index.isin(time_period)][keep]
        else:
            placebo_gaps = gaps[gaps.index.isin(time_period)]

        plt.plot(placebo_gaps, color="black", alpha=0.1)
        plt.plot(self.treated_gap, color="black", alpha=1.0)
        if treatment_time:
            plt.axvline(x=treatment_time, ymin=0.05, ymax=0.95, linestyle="dashed")
        plt.grid(grid)
        plt.show()

    def pvalue(self, treatment_time: int) -> float:
        """Calculate p-value of Abadie et al's version of Fisher's
        exact hypothesis test for no effect of treatment null, see also
        section 2.2. of :cite:`fp2018`.

        Parameters
        ----------
        treatment_time : int
            The time period that the treatment time occurred

        Returns
        -------
        float
            p-value for null hypothesis of no effect of treatment

        Raises
        ------
        ValueError
            if no placebo test has been run yet
        """
        if self.gaps is None or self.treated_gap is None:
            raise ValueError("Run a placebo test first.")

        all_ = pd.concat([self.gaps, self.treated_gap], axis=1)

        denom = all_.loc[:treatment_time].pow(2).sum(axis=0)
        num = all_.loc[treatment_time:].pow(2).sum(axis=0)

        t, _ = self.gaps.shape
        t0, _ = self.gaps.loc[:treatment_time].shape

        rmspe = (num / (t - t0)) / (denom / t0)
        return sum(
            rmspe.drop(index=self.treated_gap.name) >= rmspe.loc[self.treated_gap.name]
        ) / len(rmspe)



import unittest
import pandas as pd

from pysyncon import Dataprep, Synth


class TestSynthTexas(unittest.TestCase):
    def setUp(self):
        df = pd.read_csv("./data/texas.csv")
        self.dataprep = Dataprep(
            foo=df,
            predictors=["income", "ur", "poverty"],
            predictors_op="mean",
            time_predictors_prior=range(1985, 1994),
            special_predictors=[
                ("bmprison", [1988], "mean"),
                ("bmprison", [1990], "mean"),
                ("bmprison", [1991], "mean"),
                ("bmprison", [1992], "mean"),
                ("alcohol", [1990], "mean"),
                ("aidscapita", [1990], "mean"),
                ("aidscapita", [1991], "mean"),
                ("black", [1990], "mean"),
                ("black", [1991], "mean"),
                ("black", [1992], "mean"),
                ("perc1519", [1990], "mean"),
            ],
            dependent="bmprison",
            unit_variable="state",
            time_variable="year",
            treatment_identifier="Texas",
            controls_identifier=[
                "Alabama",
                "Alaska",
                "Arizona",
                "Arkansas",
                "California",
                "Colorado",
                "Connecticut",
                "Delaware",
                "District of Columbia",
                "Florida",
                "Georgia",
                "Hawaii",
                "Idaho",
                "Illinois",
                "Indiana",
                "Iowa",
                "Kansas",
                "Kentucky",
                "Louisiana",
                "Maine",
                "Maryland",
                "Massachusetts",
                "Michigan",
                "Minnesota",
                "Mississippi",
                "Missouri",
                "Montana",
                "Nebraska",
                "Nevada",
                "New Hampshire",
                "New Jersey",
                "New Mexico",
                "New York",
                "North Carolina",
                "North Dakota",
                "Ohio",
                "Oklahoma",
                "Oregon",
                "Pennsylvania",
                "Rhode Island",
                "South Carolina",
                "South Dakota",
                "Tennessee",
                "Utah",
                "Vermont",
                "Virginia",
                "Washington",
                "West Virginia",
                "Wisconsin",
                "Wyoming",
            ],
            time_optimize_ssr=range(1985, 1994),
        )
        self.optim_method = "BFGS"
        self.optim_initial = "ols"
        self.weights = {
            "Alabama": 0.0,
            "Alaska": 0.0,
            "Arizona": 0.0,
            "Arkansas": 0.0,
            "California": 0.407651414,
            "Colorado": 0.0,
            "Connecticut": 0.0,
            "Delaware": 0.0,
            "District of Columbia": 0.0,
            "Florida": 0.110543548,
            "Georgia": 0.0,
            "Hawaii": 0.0,
            "Idaho": 0.0,
            "Illinois": 0.36027434,
            "Indiana": 0.0,
            "Iowa": 0.0,
            "Kansas": 0.0,
            "Kentucky": 0.0,
            "Louisiana": 0.121530698,
            "Maine": 0.0,
            "Maryland": 0.0,
            "Massachusetts": 0.0,
            "Michigan": 0.0,
            "Minnesota": 0.0,
            "Mississippi": 0.0,
            "Missouri": 0.0,
            "Montana": 0.0,
            "Nebraska": 0.0,
            "Nevada": 0.0,
            "New Hampshire": 0.0,
            "New Jersey": 0.0,
            "New Mexico": 0.0,
            "New York": 0.0,
            "North Carolina": 0.0,
            "North Dakota": 0.0,
            "Ohio": 0.0,
            "Oklahoma": 0.0,
            "Oregon": 0.0,
            "Pennsylvania": 0.0,
            "Rhode Island": 0.0,
            "South Carolina": 0.0,
            "South Dakota": 0.0,
            "Tennessee": 0.0,
            "Utah": 0.0,
            "Vermont": 0.0,
            "Virginia": 0.0,
            "Washington": 0.0,
            "West Virginia": 0.0,
            "Wisconsin": 0.0,
            "Wyoming": 0.0,
        }
        self.att = {"att": 20339.375838131393, "se": 3190.4946788704715}
        self.att_time_period = range(1993, 2001)

    def test_weights(self):
        synth = Synth()
        synth.fit(
            dataprep=self.dataprep,
            optim_method=self.optim_method,
            optim_initial=self.optim_initial,
        )
        weights = pd.Series(self.weights, name="weights")
        # Allow a tolerance of 2.5%
        pd.testing.assert_series_equal(
            weights, synth.weights(round=9), check_exact=False, atol=0.025
        )

    def test_att(self):
        synth = Synth()
        synth.fit(
            dataprep=self.dataprep,
            optim_method=self.optim_method,
            optim_initial=self.optim_initial,
        )
        synth_att = synth.att(time_period=self.att_time_period)

        # Allow a tolerance of 2.5%
        att_perc_delta = abs(1.0 - self.att["att"] / synth_att["att"])
        self.assertLessEqual(att_perc_delta, 0.025)

        # Allow a tolerance of 2.5%
        se_perc_delta = abs(1.0 - self.att["se"] / synth_att["se"])
        self.assertLessEqual(se_perc_delta, 0.025)



import unittest
import numpy as np
import pandas as pd

import pysyncon


class TestPenalizedSynth(unittest.TestCase):
    def setUp(self):
        self.foo = pd.DataFrame(
            {
                "time": [1, 2, 3, 4, 1, 2, 3, 4, 1, 2, 3, 4],
                "name": [1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3],
                "dependent": np.random.random(12),
                "predictor1": np.random.random(12),
                "predictor2": np.random.random(12),
            }
        )
        self.predictors = ["predictor1"]
        self.predictors_op = "mean"
        self.dependent = "dependent"
        self.unit_variable = "name"
        self.time_variable = "time"
        self.treatment_identifier = 1
        self.treatment_identifier_list = [1, 2]
        self.controls_identifier = [2, 3]
        self.controls_identifier_alt = [3]
        self.time_predictors_prior = [2, 3]
        self.time_optimize_ssr = [1, 2, 3]
        self.special_predictors = [
            ("predictor1", [2], "mean"),
            ("predictor2", [1, 2], "median"),
            ("predictor2", [1, 2], "std"),
        ]
        self.custom_V = np.full(4, 1.0)

    def test_fit_treated(self):
        kwargs = {
            "foo": self.foo,
            "predictors": self.predictors,
            "predictors_op": self.predictors_op,
            "dependent": self.dependent,
            "unit_variable": self.unit_variable,
            "time_variable": self.time_variable,
            "time_predictors_prior": self.time_predictors_prior,
            "time_optimize_ssr": self.time_optimize_ssr,
            "special_predictors": self.special_predictors,
        }

        dataprep = pysyncon.Dataprep(
            treatment_identifier=self.treatment_identifier_list,
            controls_identifier=self.controls_identifier_alt,
            **kwargs,
        )
        pen = pysyncon.PenalizedSynth()
        self.assertRaises(ValueError, pen.fit, dataprep)

        dataprep = pysyncon.Dataprep(
            treatment_identifier=self.treatment_identifier,
            controls_identifier=self.controls_identifier,
            **kwargs,
        )
        pen = pysyncon.PenalizedSynth()
        try:
            pen.fit(dataprep)
        except Exception as e:
            self.fail(f"PenalizedSynth fit with single treated failed: {e}.")

        dataprep = pysyncon.Dataprep(
            treatment_identifier=[self.treatment_identifier],
            controls_identifier=self.controls_identifier,
            **kwargs,
        )
        pen = pysyncon.PenalizedSynth()
        try:
            pen.fit(dataprep)
        except Exception as e:
            self.fail(f"PenalizedSynth fit with single treated in list failed: {e}.")

    def test_X0_X1_fit(self):
        pen = pysyncon.PenalizedSynth()

        # X1 needs to be pd.Series
        X0 = pd.DataFrame(np.random.rand(5, 5))
        X1 = pd.DataFrame(np.random.rand(5, 2))
        self.assertRaises(TypeError, pen.fit, X0=X0, X1=X1)

        # X1 needs to be pd.Series
        X0 = pd.DataFrame(np.random.rand(5, 5))
        X1 = pd.DataFrame(np.random.rand(5, 1))
        self.assertRaises(TypeError, pen.fit, X0=X0, X1=X1)

    def test_fit_no_data(self):
        pen = pysyncon.PenalizedSynth()
        self.assertRaises(ValueError, pen.fit)

    def test_fit_custom_V(self):
        kwargs = {
            "foo": self.foo,
            "predictors": self.predictors,
            "predictors_op": self.predictors_op,
            "dependent": self.dependent,
            "unit_variable": self.unit_variable,
            "time_variable": self.time_variable,
            "treatment_identifier": self.treatment_identifier,
            "controls_identifier": self.controls_identifier,
            "time_predictors_prior": self.time_predictors_prior,
            "time_optimize_ssr": self.time_optimize_ssr,
            "special_predictors": self.special_predictors,
        }

        dataprep = pysyncon.Dataprep(**kwargs)
        pen = pysyncon.PenalizedSynth()
        try:
            pen.fit(dataprep=dataprep, custom_V=self.custom_V)
        except Exception as e:
            self.fail(f"PenalizedSynth fit failed with custom_V: {e}")



import unittest
from unittest.mock import patch, Mock
import numpy as np
import pandas as pd

from pysyncon import Dataprep, Synth
from pysyncon.utils import HoldoutSplitter, CrossValidationResult, PlaceboTest


class TestHoldoutSplitter(unittest.TestCase):
    def test_values(self):
        cases = [(3, 3, 1), (3, 3, 2), (5, 1, 1), (5, 1, 2)]
        for case in cases:
            with self.subTest(case=case):
                rows, columns, holdout = case
                df = pd.DataFrame(np.random.random(size=(rows, columns)))
                ser = pd.Series(np.random.random(size=rows))

                iter_len = 0
                for df_, df_h, ser_, ser_h in HoldoutSplitter(
                    df=df, ser=ser, holdout_len=holdout
                ):
                    self.assertIsInstance(df_, pd.DataFrame)
                    pd.testing.assert_frame_equal(
                        df_,
                        df.drop(index=df.index[iter_len : iter_len + holdout,]),
                    )

                    self.assertIsInstance(ser_, pd.Series)
                    pd.testing.assert_series_equal(
                        ser_,
                        ser.drop(index=ser.index[iter_len : iter_len + holdout]),
                    )

                    self.assertIsInstance(df_h, pd.DataFrame)
                    pd.testing.assert_frame_equal(
                        df_h,
                        df.iloc[iter_len : iter_len + holdout,],
                    )

                    self.assertIsInstance(ser_h, pd.Series)
                    pd.testing.assert_series_equal(
                        ser_h,
                        ser.iloc[iter_len : iter_len + holdout],
                    )
                    iter_len += 1
                self.assertEqual(iter_len - 1, rows - holdout)

    def test_errs(self):
        cases = [(1, 1, 2, 2), (2, 2, 1, 1), (3, 2, 1, 2), (2, 1, 2, 3)]
        for case in cases:
            with self.subTest(case=case):
                df_rows, df_cols, holdout, ser_rows = case

                df = pd.DataFrame(np.random.random(size=(df_rows, df_cols)))
                ser = pd.Series(np.random.random(size=ser_rows))

                self.assertRaises(
                    ValueError,
                    HoldoutSplitter,
                    df=df,
                    ser=ser,
                    holdout_len=holdout,
                )

        cases = [(1, 1, 0, 1), (2, 2, 2, 2), (3, 3, 4, 3)]
        for case in cases:
            with self.subTest(case=case):
                df_rows, df_cols, holdout, ser_rows = case

                df = pd.DataFrame(np.random.random(size=(df_rows, df_cols)))
                ser = pd.Series(np.random.random(size=ser_rows))

                self.assertRaises(
                    ValueError,
                    HoldoutSplitter,
                    df=df,
                    ser=ser,
                    holdout_len=holdout,
                )


class TestCrossValidationResult(unittest.TestCase):
    def test_best_lambda(self):
        cases = [1, 2, 3, 10]
        for case in cases:
            with self.subTest(case=case):
                cv_result = CrossValidationResult(
                    lambdas=np.random.random(size=case),
                    errors_mean=np.random.random(size=case),
                    errors_se=np.random.random(size=case),
                )

                best_lambda = cv_result.best_lambda()
                min_mean = cv_result.errors_mean.min()
                min_mean_idx = cv_result.errors_mean.argmin()
                min_mean_se = cv_result.errors_se[min_mean_idx]
                self.assertEqual(
                    best_lambda,
                    cv_result.lambdas[cv_result.errors_mean <= min_mean + min_mean_se]
                    .max()
                    .item(),
                )

                best_lambda = cv_result.best_lambda(min_1se=False)
                min_mean_idx = cv_result.errors_mean.argmin()
                self.assertEqual(best_lambda, cv_result.lambdas[min_mean_idx].item())

    @patch("pysyncon.utils.plt")
    def test_result_plot(self, mock_plt: Mock):
        cv_result = CrossValidationResult(
            lambdas=np.random.random(size=10),
            errors_mean=np.random.random(size=10),
            errors_se=np.random.random(size=10),
        )
        cv_result.plot()

        self.assertEqual(mock_plt.errorbar.call_count, 1)
        _, kwargs = mock_plt.errorbar.call_args
        self.assertEqual(kwargs["ecolor"], "black")
        self.assertEqual(kwargs["capsize"], 2)

        mock_plt.xlabel.assert_called_with("Lambda")
        mock_plt.ylabel.assert_called_with("Mean error")
        mock_plt.xscale.assert_called_with("log")
        mock_plt.yscale.assert_called_with("log")
        mock_plt.title.assert_called_with("Cross validation result")
        mock_plt.grid.assert_called()
        mock_plt.show.assert_called()


class TestPlaceboTests(unittest.TestCase):
    def setUp(self):
        # 1 -> treated, (2, 3) -> controls
        self.dataprep = Dataprep(
            foo=pd.DataFrame(
                {
                    "time": [1, 2, 3, 4, 1, 2, 3, 4, 1, 2, 3, 4],
                    "name": [1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3],
                    "dependent": np.random.random(12),
                    "predictor1": np.random.random(12),
                    "predictor2": np.random.random(12),
                }
            ),
            predictors=["predictor1"],
            predictors_op="mean",
            dependent="dependent",
            unit_variable="name",
            time_variable="time",
            treatment_identifier=1,
            controls_identifier=[2, 3],
            time_predictors_prior=[2, 3],
            time_optimize_ssr=[1, 2, 3],
            special_predictors=[
                ("predictor1", [2], "mean"),
                ("predictor2", [1, 2], "median"),
                ("predictor2", [1, 2], "std"),
            ],
        )
        self.synth = Synth()
        self.synth.fit(dataprep=self.dataprep)

        self.placebo_test = PlaceboTest()
        self.placebo_test.fit(dataprep=self.dataprep, scm=self.synth)

    @patch("pysyncon.utils.plt")
    def test_gaps_plot(self, mock_plt: Mock):
        self.placebo_test.gaps_plot()

        self.assertEqual(mock_plt.plot.call_count, 2)
        _, kwargs = mock_plt.plot.call_args
        self.assertEqual(kwargs["color"], "black")
        self.assertEqual(kwargs["alpha"], 1.0)
        mock_plt.axvline.assert_not_called()
        mock_plt.grid.assert_called()

    @patch("pysyncon.utils.plt")
    def test_gaps_plot_axvline(self, mock_plt: Mock):
        self.placebo_test.gaps_plot(treatment_time=3)

        mock_plt.axvline.assert_called()
        _, kwargs = mock_plt.axvline.call_args
        self.assertEqual(kwargs["ymin"], 0.05)
        self.assertEqual(kwargs["ymax"], 0.95)
        self.assertEqual(kwargs["linestyle"], "dashed")

    @patch("pysyncon.utils.plt")
    def test_gaps_plot_mspe_threshold(self, mock_plt: Mock):
        self.placebo_test.gaps_plot(treatment_time=3, mspe_threshold=1)



from __future__ import annotations
from typing import Optional

import numpy as np
import pandas as pd

from .dataprep import Dataprep
from .base import BaseSynth, VanillaOptimMixin
from .utils import HoldoutSplitter, CrossValidationResult


class AugSynth(BaseSynth, VanillaOptimMixin):
    """Implementation of the augmented synthetic control method due to Ben-
    Michael, Feller & Rothstein :cite:`augsynth2021`.

    The implementation follows the augsynth R package with the option
    `progfunc="Ridge"`.
    """

    def __init__(self) -> None:
        super().__init__()
        self.lambda_: Optional[float] = None
        self.cv_result: Optional[CrossValidationResult] = None

    def fit(self, dataprep: Dataprep, lambda_: Optional[float] = None) -> None:
        """Fit the model/calculate the weights.

        Parameters
        ----------
        dataprep : Dataprep, optional
            :class:`Dataprep` object containing data to model.
        lambda_ : float, optional
            Ridge parameter to use. If not supplied, then it is obtained by
            cross-validation, by default None
        """
        if (
            isinstance(dataprep.treatment_identifier, (list, tuple))
            and len(dataprep.treatment_identifier) > 1
        ):
            raise ValueError("AugSynth requires exactly one treated unit.")
        self.dataprep = dataprep
        Z0, Z1 = self.dataprep.make_covariate_mats()
        X0, X1 = self.dataprep.make_outcome_mats()

        X0_demean, X1_demean, Z0_normal, Z1_normal = self._normalize(X0, X1, Z0, Z1)
        X0_stacked = pd.concat([X0_demean, Z0_normal], axis=0)
        X1_stacked = pd.concat([X1_demean, Z1_normal], axis=0)

        if lambda_ is None:
            lambdas = self.generate_lambdas(X0)
            self.cv_result = self.cross_validate(X0, X1, lambdas)
            self.lambda_ = self.cv_result.best_lambda()
        else:
            self.lambda_ = lambda_

        n_r, _ = X0.shape
        V_mat = np.diag(np.full(n_r, 1 / n_r))
        W, _ = self.w_optimize(V_mat=V_mat, X0=X0.to_numpy(), X1=X1.to_numpy())

        W_ridge = self.solve_ridge(
            X1_stacked.to_numpy(), X0_stacked.to_numpy(), W, self.lambda_
        )
        self.W = W + W_ridge

    @staticmethod
    def solve_ridge(
        A: np.ndarray, B: np.ndarray, W: np.ndarray, lambda_: float
    ) -> np.ndarray:
        """Calculate the ridge adjustment to the weights.

        :meta private:
        """
        M = A - B @ W
        N = np.linalg.inv(B @ B.T + lambda_ * np.identity(B.shape[0]))
        return M @ N @ B

    def _normalize(
        self, X0: pd.DataFrame, X1: pd.Series, Z0: pd.DataFrame, Z1: pd.Series
    ) -> tuple[pd.DataFrame, pd.Series, pd.DataFrame, pd.Series]:
        """Normalise the data before the weight calculation.

        :meta private:
        """
        X0_demean = X0.subtract(X0.mean(axis=1), axis=0)
        X1_demean = X1.subtract(X0.mean(axis=1), axis=0)

        Z0_demean = Z0.subtract(Z0.mean(axis=1), axis=0)
        Z1_demean = Z1.subtract(Z0.mean(axis=1), axis=0)

        Z0_std = Z0_demean.std(axis=1)
        X0_std = X0_demean.to_numpy().std(ddof=1).item()

        Z0_normal = Z0_demean.divide(Z0_std, axis=0) * X0_std
        Z1_normal = Z1_demean.divide(Z0_std, axis=0) * X0_std
        return X0_demean, X1_demean, Z0_normal, Z1_normal

    def cross_validate(
        self, X0: np.ndarray, X1: np.ndarray, lambdas: np.ndarray, holdout_len: int = 1
    ) -> CrossValidationResult:
        """Method that calculates the mean error and standard error to the mean
        error using a cross-validation procedure for the given ridge parameter
        values.

        :meta private:
        """
        V = np.identity(X0.shape[0] - holdout_len)
        res = list()
        for X0_t, X0_v, X1_t, X1_v in HoldoutSplitter(X0, X1, holdout_len=holdout_len):
            W, _ = self.w_optimize(V_mat=V, X0=X0_t.to_numpy(), X1=X1_t.to_numpy())
            this_res = list()
            for l in lambdas:
                ridge_weights = self.solve_ridge(A=X1_t, B=X0_t, W=W, lambda_=l)
                W_aug = W + ridge_weights
                err = (X1_v - X0_v @ W_aug).pow(2).sum()
                this_res.append(err.item())
            res.append(this_res)
        means = np.array(res).mean(axis=0)
        ses = np.array(res).std(axis=0) / np.sqrt(len(lambdas))
        return CrossValidationResult(lambdas, means, ses)

    def generate_lambdas(
        self, X: pd.DataFrame, lambda_min_ratio: float = 1e-8, n_lambda: int = 20
    ) -> np.ndarray:
        """Generate a suitable set of lambdas to run the cross-validation
        procedure on.

        :meta private:
        """
        _, sing, _ = np.linalg.svd(X.T)
        lambda_max = sing[0].item() ** 2.0
        scaler = lambda_min_ratio ** (1 / n_lambda)
        return lambda_max * (scaler ** np.array(range(n_lambda)))



from __future__ import annotations
from typing import Optional, Callable, Literal

import numpy as np
import pandas as pd

from pysyncon.base import BaseSynth


class ConformalInference:
    """Implementation of the conformal inference based confidence intervals
    following Chernozhukov et al. :cite:`inference2021`
    """

    def __init__(self) -> None:
        pass

    def confidence_intervals(
        self,
        alpha: float,
        scm: BaseSynth,
        Z0: pd.DataFrame,
        Z1: pd.Series,
        pre_periods: list,
        post_periods: list,
        tol: float = 0.1,
        max_iter: int = 50,
        step_sz: Optional[float] = None,
        step_sz_div: float = 20.0,
        verbose: bool = True,
        scm_fit_args: dict = {},
    ) -> pd.DataFrame:
        """Confidence intervals obtained from test-inversion, where
        the p-values are obtained by adjusted re-fits of the data
        following Chernozhukov et al. :cite:`inference2021`.

        Parameters
        ----------
        alpha : float
            The required significance level, e.g. alpha = 0.05 will
            yield a confidence level of 100 * (1 - alpha) = 95%.
        scm : BaseSynth
            The synth object to calculate the confidence interval for.
        Z0 : pandas.DataFrame, shape (n, c)
            A matrix of the time series of the outcome variable with each
            column corresponding to a control unit and the rows are the time
            steps.
        Z1 : pd.Series
            Column vector giving the outcome variable values over time for the
            treated unit.
        tol : float
            The required tolerance (accuracy) required when calculating the
            lower/upper cut-off point of the confidence interval. The search
            will try to obtain this tolerance level but will not exceed `max_iter`
            iterations trying to achieve that.
        pre_periods : list
            The time-periods to use for the optimization when refitting the
            data with the adjusted outcomes.
        post_periods : list
            The time-periods to calculate confidence intervals for.
        max_iter : int, optional
            Maximum number of times to re-fit the data when trying to locate
            the lower/upper cut-off point, by default 50
        step_sz : Optional[float], optional
            Step size to use when searching for an interval that contains the
            lower or upper cut-off point of the confidence interval, by default None
        step_sz_div : float, optional
            Alternative way to define step size: it is the fraction that defines
            step-size in terms of the standard deviation of the att, i.e. if
            `step_sz_div=20.0` then the step size used will be (att +/- 2.5 * std(att)) / 20.0,
            by default 20.0
        verbose : bool, optional
            Print output, by default True
        scm_fit_args : dict, optional
            A dictionary defining anything extra that should be provided to the
            synthetic control object `fit` method when doing the refits, by default {}

        Returns
        -------
        pd.DataFrame
            A pandas.DataFrame indexed by `post_periods`, with 3 columns: `value` that
            gives the calculated treatment effect, `lower_ci` that gives the value
            defining the lower-end of the confidence interval, `upper_ci` that gives
            the value defining the upper-end of the confidence interval.

        Raises
        ------
        TypeError
            if `alpha` is not a float
        ValueError
            if `alpha` is not in the open interval (0, 1).
        TypeError
            if `max_iter` is not an integer
        ValueError
            if `max_iter` is not at least 1
        TypeError
            if `tol` is not a float
        ValueError
            if `tol` is less than 0.0
        TypeError
            if `step_sz` is not a float
        ValueError
            if `step_sz` is not greater than 0.0
        TypeError
            if `step_sz_div` is not a float
        ValueError
            if `step_sz_div` is not greater than 0.0
        """
        if not isinstance(alpha, float):
            raise TypeError("`alpha` must be a float")
        elif not 0.0 < alpha < 1.0:
            raise ValueError("`alpha` must be greater than 0.0 and less than 1.0")
        if not isinstance(max_iter, int):
            raise TypeError("`max_iter` must be an integer")
        elif max_iter < 1:
            raise ValueError("`max_iter` must be at least 1")
        if not isinstance(tol, float):
            raise TypeError("`tol` must be a float")
        elif tol <= 0.0:
            raise ValueError("`tol` must be greater than 0.0")
        if step_sz != None:
            if not isinstance(step_sz, float):
                raise TypeError("`step_sz` should be a float")
            elif step_sz <= 0.0:
                raise ValueError("`step_sz` should be greater than 0.0")
            elif step_sz <= tol:
                raise ValueError("`step_sz` must be greater than `tol`.")
        if not isinstance(step_sz_div, float):
            raise TypeError("`step_sz_div` must be a float")
        elif step_sz_div <= 0.0:
            raise ValueError("`step_sz_div` must be greater than 0.0")
        if scm.W is None:
            raise ValueError("No weight matrix available; fit data first.")

        gaps = scm._gaps(Z0=Z0, Z1=Z1)
        if step_sz is None:
            # Try to guess a step-size
            if len(post_periods) > 1:
                factor = np.std(gaps.loc[post_periods])
            else:
                factor = gaps.loc[post_periods].item() / 2.0
            step_sz = 2.5 * factor / step_sz_div
            if step_sz <= tol:
                # Failed to guess a sensible step-size :(
                step_sz = 1.1 * tol

        conf_interval = dict()
        n_periods = len(post_periods)
        for idx, post_period in enumerate(post_periods, 1):
            if verbose:
                print(
                    f"({idx}/{n_periods}) Calculating confidence interval "
                    f"for time-period t={post_period}..."
                )
            new_time_range = pre_periods + [post_period]
            Z0_new, Z1_new = Z0.loc[new_time_range], Z1.loc[new_time_range]
            Z1_post_orig = Z1_new.loc[post_period].item()

            def _compute_p_value(g):
                Z1_new.loc[post_period] = Z1_post_orig - g
                scm.fit(Z0=Z0_new, Z1=Z1_new, **scm_fit_args)
                _gaps = scm._gaps(Z0=Z0_new, Z1=Z1_new)

                u_hat = _gaps.loc[new_time_range]
                u_hat_post = u_hat.loc[post_period]
                return np.mean(abs(u_hat) >= abs(u_hat_post))

            lower_ci = self._root_search(
                fn=lambda x: _compute_p_value(x) - alpha,
                x0=gaps.loc[post_period],
                direction=-1.0,
                tol=tol,
                step_sz=step_sz,
                max_iter=max_iter,
            )

            upper_ci = self._root_search(
                fn=lambda x: _compute_p_value(x) - alpha,
                x0=gaps.loc[post_period],
                direction=1.0,
                tol=tol,
                step_sz=step_sz,
                max_iter=max_iter,
            )

            conf_interval[post_period] = (lower_ci, upper_ci)
            if verbose:
                print(
                    f"\t{100 * (1 - alpha)}% CI: [{round(lower_ci, 3)}, {round(upper_ci, 3)}]"
                )

        df_ci = pd.DataFrame.from_dict(
            conf_interval, orient="index", columns=["lower_ci", "upper_ci"]
        )
        df_ci = pd.concat([gaps.loc[post_periods].rename("value"), df_ci], axis=1)
        df_ci.index.name = "time"
        return df_ci

    def _root_search(
        self,
        fn: Callable,
        x0: float,
        direction: Literal[+1, -1],
        tol: float,
        step_sz: float,
        max_iter: int,
        theta: float = 0.75,
        phi: float = 1.3,
    ) -> float:
        """Search for a root

        Parameters
        ----------
        fn : callable
            Function to find a root of
        x0 : float
            Starting point
        direction : int
            Direction, either -1.0 or +1.0.
        tol : float
            Tolerance
        step_sz : float
            Step size in the search
        max_iter : int
            Maximum number of iterations
        theta : float, optional
            Step size reduction factor, should be positive and < 1.0, by default 0.75
        phi : float, optional
            Step size increase factor, should be positive and > 1.0, by default 1.3

        Returns
        -------
        float
            Root of the function

        Raises
        ------
        Exception
            if `max_iter` iterations exceeded before satisfying tolerance condition.

        :meta private:
        """
        x, gamma = x0, step_sz
        for _ in range(max_iter):
            if gamma <= tol:
                return x
            y = fn(x + gamma * direction)
            if y > 0.0:
                x = x + gamma * direction
                gamma = phi * gamma
            else:
                gamma = theta * gamma
        raise Exception(
            "Exceeded `max_iter` iterations without satisfying tolerance requirement."
        )



from __future__ import annotations
from typing import Optional, Literal, Sequence
from abc import ABCMeta, abstractmethod

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy.optimize import minimize, Bounds, LinearConstraint

from .dataprep import Dataprep, IsinArg_t


class BaseSynth(metaclass=ABCMeta):
    """Metaclass that defines methods common to synthetic control methods."""

    def __init__(self) -> None:
        self.dataprep: Optional[Dataprep] = None
        self.W: Optional[np.ndarray] = None
        self.W_names: Optional[Sequence] = None

    @abstractmethod
    def fit(*args, **kwargs) -> None:
        raise NotImplementedError

    def _synthetic(self, Z0: pd.DataFrame) -> pd.Series:
        """Assemble the synthetic unit using the calculated weight matrix.

        Parameters
        ----------
        Z0 : pandas.DataFrame, shape (n, c)
            A matrix of the time series of the outcome variable with each
            column corresponding to a control unit and the rows are the time
            steps

        Returns
        -------
        pd.Series
            Time series of the synthetic unit.
        """
        ts_synthetic = (Z0 * self.W).sum(axis=1)
        return ts_synthetic

    def path_plot(
        self,
        time_period: Optional[IsinArg_t] = None,
        treatment_time: Optional[int] = None,
        grid: bool = True,
        Z0: Optional[pd.DataFrame] = None,
        Z1: Optional[pd.Series] = None,
    ) -> None:
        """Plot the outcome variable over time for the treated unit and the
        synthetic control.

        Parameters
        ----------
        time_period : Iterable | pandas.Series | dict, optional
            Time range to plot, if none is supplied then the time range used
            is the time period over which the optimisation happens, by default
            None
        treatment_time : int, optional
            If supplied, plot a vertical line at the time period that the
            treatment time occurred, by default None
        grid : bool, optional
            Whether or not to plot a grid, by default True
        Z0 : pandas.DataFrame, shape (n, c), optional
            The matrix of the time series of the outcome variable for the control units.
            If no dataprep is set, then this must be supplied along with Z1, by default None.
        Z1 : pandas.Series, shape (n, 1), optional
            The matrix of the time series of the outcome variable for the treated unit.
            If no dataprep is set, then this must be supplied along with Z0, by default None.

        Raises
        ------
        ValueError
            If there is no weight matrix available
        ValueError
            If there is no :class:`Dataprep` object set or (Z0, Z1) is not supplied
        """
        if self.dataprep is not None:
            Z0, Z1 = self.dataprep.make_outcome_mats(time_period=time_period)
        elif Z0 is None or Z1 is None:
            raise ValueError("dataprep must be set or (Z0, Z1) must be set for plots.")
        if self.W is None:
            raise ValueError("No weight matrix available; fit data first.")

        ts_synthetic = self._synthetic(Z0=Z0)
        plt.plot(Z1, color="black", linewidth=1, label=Z1.name)
        plt.plot(
            ts_synthetic,
            color="black",
            linewidth=1,
            linestyle="dashed",
            label="Synthetic",
        )
        if self.dataprep is not None:
            plt.ylabel(self.dataprep.dependent)
        if treatment_time:
            plt.axvline(x=treatment_time, ymin=0.05, ymax=0.95, linestyle="dashed")
        plt.legend()
        plt.grid(grid)
        plt.show()

    def _gaps(self, Z0: pd.DataFrame, Z1: pd.Series) -> pd.Series:
        """Calculate the gaps (difference between factual
        and estimated counterfactual)

        Parameters
        ----------
        Z0 : pandas.DataFrame, shape (n, c)
            A matrix of the time series of the outcome variable with each
            column corresponding to a control unit and the rows are the time
            steps
        Z1 : pandas.DataFrame, shape (n, 1)
            A matrix of the time series of the outcome variable for the treated
            unit and the rows are the time steps

        Returns
        -------
        pd.Series
            Series containing the gaps

        :meta private:
        """
        ts_synthetic = self._synthetic(Z0=Z0)
        ts_gap = Z1 - ts_synthetic
        return ts_gap

    def gaps_plot(
        self,
        time_period: Optional[IsinArg_t] = None,
        treatment_time: Optional[int] = None,
        grid: bool = True,
        Z0: Optional[pd.DataFrame] = None,
        Z1: Optional[pd.Series] = None,
    ) -> None:
        """Plots the gap between the treated unit and the synthetic unit over
        time.

        Parameters
        ----------
        time_period : Iterable | pandas.Series | dict, optional
            Time range to plot, if none is supplied then the time range used
            is the time period over which the optimisation happens, by default
            None
        treatment_time : int, optional
            If supplied, plot a vertical line at the time period that the
            treatment time occurred, by default None
        grid : bool, optional
            Whether or not to plot a grid, by default True
        Z0 : pandas.DataFrame, shape (n, c), optional
            The matrix of the time series of the outcome variable for the control units.
            If no dataprep is set, then this must be supplied along with Z1, by default None.
        Z1 : pandas.Series, shape (n, 1), optional
            The matrix of the time series of the outcome variable for the treated unit.
            If no dataprep is set, then this must be supplied along with Z0, by default None.

        Raises
        ------
        ValueError
            If there is no weight matrix available
        ValueError
            If there is no :class:`Dataprep` object set or (Z0, Z1) is not supplied
        """
        if self.dataprep is not None:
            Z0, Z1 = self.dataprep.make_outcome_mats(time_period=time_period)
        elif Z0 is None or Z1 is None:
            raise ValueError("dataprep must be set or (Z0, Z1) must be set for plots.")
        if self.W is None:
            raise ValueError("No weight matrix available; fit data first.")

        ts_gap = self._gaps(Z0=Z0, Z1=Z1)
        plt.plot(ts_gap, color="black", linewidth=1)
        if self.dataprep is not None:
            plt.ylabel(self.dataprep.dependent)
        plt.hlines(
            y=0,
            xmin=min(ts_gap.index),
            xmax=max(ts_gap.index),
            color="black",
            linestyle="dashed",
        )
        if treatment_time:
            plt.axvline(x=treatment_time, ymin=0.05, ymax=0.95, linestyle="dashed")
        plt.grid(grid)
        plt.show()

    def weights(self, round: int = 3, threshold: Optional[float] = None) -> pd.Series:
        """Return a ``pandas.Series`` of the weights for each control unit.

        Parameters
        ----------
        round : int, optional
            Round the weights to given number of places, by default 3
        threshold : float, optional
            If supplied, will only show weights above this value, by default
            None

        Returns
        -------
        pandas.Series
            The weights computed

        Raises
        ------
        ValueError
            If there is no weight matrix available
        """
        if self.W is None:
            raise ValueError("No weight matrix available; fit data first.")
        if self.dataprep is None:
            weights_ser = pd.Series(self.W, index=self.W_names, name="weights")
        else:
            weights_ser = pd.Series(
                self.W, index=list(self.dataprep.controls_identifier), name="weights"
            )
        weights_ser = (
            weights_ser[weights_ser >= threshold] if threshold else weights_ser
        )
        return weights_ser.round(round)

    def summary(
        self,
        round: int = 3,
        X0: Optional[pd.DataFrame] = None,
        X1: Optional[pd.Series] = None,
    ) -> pd.DataFrame:
        """Generates a ``pandas.DataFrame`` with summary data. The
        first column will show the mean value of each predictor over the time
        period ``time_predictors_prior`` for the treated unit and the second
        column the case of the synthetic unit and finally there will be a
        column 'sample mean' that shows the mean value of each predictor
        over the time period ``time_predictors_prior`` across all the control
        units, i.e. this will be the same as a synthetic control where all
        the weights are equal.

        Parameters
        ----------
        round : int, optional
            Round the table values to the given number of places, by
            default 3
        X0 : pd.DataFrame, shape (n_cov, n_controls), optional
            Matrix with each column corresponding to a control unit and each
            row is a covariate. If no dataprep is set, then this must be
            supplied along with X1, by default None.
        X1 : pandas.Series, shape (n_cov, 1), optional
            Column vector giving the covariate values for the treated unit.
            If no dataprep is set, then this must be supplied along with Z1,
            by default None.

        Returns
        -------
        pandas.DataFrame
            Summary data.

        Raises
        ------
        ValueError
            If there is no weight matrix available
        ValueError
            If there is no :class:`Dataprep` object set or (Z0, Z1) is not supplied
        """
        if self.W is None:
            raise ValueError("No weight matrix available: fit data first.")
        if self.dataprep is not None:
            X0, X1 = self.dataprep.make_covariate_mats()
        elif X0 is None or X1 is None:
            raise ValueError(
                "dataprep must be set or (X0, X1) must be set for summary."
            )

        treated = X1.rename("treated")
        synthetic = (X0 * self.W).sum(axis=1).rename("synthetic")
        sample_mean = X0.mean(axis=1).rename("sample mean")

        return pd.concat([treated, synthetic, sample_mean], axis=1).round(round)

    def att(
        self,
        time_period: IsinArg_t,
        Z0: Optional[pd.DataFrame] = None,
        Z1: Optional[pd.Series] = None,
    ) -> dict[str, float]:
        """Computes the average treatment effect on the treated unit (ATT) and
        the standard error to the value over the chosen time-period.

        Parameters
        ----------
        time_period : Iterable | pandas.Series | dict, optional
            Time period to compute the ATT over.
        Z0 : pandas.DataFrame, shape (n, c), optional
            The matrix of the time series of the outcome variable for the control units.
            If no dataprep is set, then this must be supplied along with Z1, by default None.
        Z1 : pandas.Series, shape (n, 1), optional
            The matrix of the time series of the outcome variable for the treated unit.
            If no dataprep is set, then this must be supplied along with Z0, by default None.

        Returns
        -------
        dict
            A dictionary with the ATT value and the standard error to the ATT.

        Raises
        ------
        ValueError
            If there is no weight matrix available
        ValueError
            If there is no :class:`Dataprep` object set or (Z0, Z1) is not supplied
        """
        if self.W is None:
            raise ValueError("No weight matrix available; fit data first.")
        if self.dataprep is not None:
            Z0, Z1 = self.dataprep.make_outcome_mats(time_period=time_period)
            gaps = self._gaps(Z0=Z0, Z1=Z1)
        if Z0 is not None and Z1 is not None:
            gaps = self._gaps(Z0=Z0.loc[time_period, :], Z1=Z1.loc[time_period])
        else:
            raise ValueError("dataprep must be set or (Z0, Z1) must be set for att.")
        att = np.mean(gaps)
        se = np.std(gaps, ddof=1) / np.sqrt(len(time_period))

        return {"att": att.item(), "se": se.item()}

    def mspe(
        self, Z0: Optional[pd.DataFrame] = None, Z1: Optional[pd.Series] = None
    ) -> float:
        """Returns the mean square prediction error in the fit of
        the synthetic control versus the treated unit over the
        optimization time-period.

        Parameters
        ----------
        Z0 : pandas.DataFrame, shape (n, c), optional
            The matrix of the time series of the outcome variable for the control units.
            If no dataprep is set, then this must be supplied along with Z1, by default None.
        Z1 : pandas.Series, shape (n, 1), optional
            The matrix of the time series of the outcome variable for the treated unit.
            If no dataprep is set, then this must be supplied along with Z0, by default None.

        Returns
        -------
        float
            Mean square prediction Error

        Raises
        ------
        ValueError
            If the fit method has not been run (no weights available.)
        ValueError
            If there is no :class:`Dataprep` object set or (Z0, Z1) is not supplied
        """
        if self.W is None:
            raise ValueError("No weight matrix available; fit data first.")
        if self.dataprep is not None:
            Z0, Z1 = self.dataprep.make_outcome_mats(
                time_period=self.dataprep.time_optimize_ssr
            )
        if Z0 is None or Z1 is None:
            raise ValueError("dataprep must be set or (Z0, Z1) must be set for plots.")
        ts_synthetic = self._synthetic(Z0=Z0)

        n = len(ts_synthetic)
        return (1 / n) * (Z1 - ts_synthetic).pow(2).sum().item()

    def mape(
        self, Z0: Optional[pd.DataFrame] = None, Z1: Optional[pd.Series] = None
    ) -> float:
        """Returns the mean absolute percentage error in the fit of
        the synthetic control versus the treated unit over the
        optimization time-period.

        Parameters
        ----------
        Z0 : pandas.DataFrame, shape (n, c), optional
            The matrix of the time series of the outcome variable for the control units.
            If no dataprep is set, then this must be supplied along with Z1, by default None.
        Z1 : pandas.Series, shape (n, 1), optional
            The matrix of the time series of the outcome variable for the treated unit.
            If no dataprep is set, then this must be supplied along with Z0, by default None.

        Returns
        -------
        float
            Mean absolute percentage error

        Raises
        ------
        ValueError
            If the fit method has not been run (no weights available.)
        ValueError
            If there is no :class:`Dataprep` object set or (Z0, Z1) is not supplied
        """
        if self.W is None:
            raise ValueError("No weight matrix available; fit data first.")
        if self.dataprep is not None:
            Z0, Z1 = self.dataprep.make_outcome_mats(
                time_period=self.dataprep.time_optimize_ssr
            )
        if Z0 is None or Z1 is None:
            raise ValueError("dataprep must be set or (Z0, Z1) must be set for plots.")
        ts_synthetic = self._synthetic(Z0=Z0)

        n = len(ts_synthetic)
        return (1 / n) * ((Z1 - ts_synthetic) / Z1).abs().sum().item()

    def mae(
        self, Z0: Optional[pd.DataFrame] = None, Z1: Optional[pd.Series] = None
    ) -> float:
        """Returns the mean absolute error in the fit of
        the synthetic control versus the treated unit over the
        optimization time-period.

        Parameters
        ----------
        Z0 : pandas.DataFrame, shape (n, c), optional
            The matrix of the time series of the outcome variable for the control units.
            If no dataprep is set, then this must be supplied along with Z1, by default None.
        Z1 : pandas.Series, shape (n, 1), optional
            The matrix of the time series of the outcome variable for the treated unit.
            If no dataprep is set, then this must be supplied along with Z0, by default None.

        Returns
        -------
        float
            Mean absolute error

        Raises
        ------
        ValueError
            If the fit method has not been run (no weights available.)
        ValueError
            If there is no :class:`Dataprep` object set or (Z0, Z1) is not supplied
        """
        if self.W is None:
            raise ValueError("No weight matrix available; fit data first.")
        if self.dataprep is not None:
            Z0, Z1 = self.dataprep.make_outcome_mats(
                time_period=self.dataprep.time_optimize_ssr
            )
        if Z0 is None or Z1 is None:
            raise ValueError("dataprep must be set or (Z0, Z1) must be set for plots.")
        ts_synthetic = self._synthetic(Z0=Z0)

        n = len(ts_synthetic)
        return (1 / n) * (Z1 - ts_synthetic).abs().sum().item()


class VanillaOptimMixin:
    @staticmethod
    def w_optimize(
        V_mat: np.ndarray,
        X0: np.ndarray,
        X1: np.ndarray,
        qp_method: Literal["SLSQP"] = "SLSQP",
        qp_options: dict = {"maxiter": 1000},
    ) -> tuple[np.ndarray, float]:
        """Solves the inner part of the quadratic minimization problem for a
        given V matrix (see Abadie and Gardeazabal :cite:`basque2003`).

        Parameters
        ----------
        V_mat : numpy.ndarray, shape (c, c)
            V matrix using the notation of the Abadie, Diamond & Hainmueller
            paper defining.
        X0 : numpy.ndarray, shape (m, c)
            Matrix with each column corresponding to a control unit and each
            row is covariates.
        X1 : numpy.ndarray, shape (m,)
            Column vector giving the covariate values for the treated unit.
        qp_method : str, optional
            Minimization routine to use in scipy minimize to solve the problem
            , by default "SLSQP"
        qp_options : dict, optional
            Options for scipy minimize, by default {"maxiter": 1000}

        Returns
        -------
        tuple[np.ndarray, float]
            tuple of the optimal weights and the loss

        :meta private:
        """
        _, n_c = X0.shape

        P = X0.T @ V_mat @ X0
        q = X1.T @ V_mat @ X0

        def fun(x):
            return 0.5 * x.T @ P @ x - q.T @ x

        bounds = Bounds(lb=np.full(n_c, 0.0), ub=np.full(n_c, 1.0))
        constraints = LinearConstraint(A=np.full(n_c, 1.0), lb=1.0, ub=1.0)

        x0 = np.full(n_c, 1 / n_c)
        res = minimize(
            fun=fun,
            x0=x0,
            bounds=bounds,
            constraints=constraints,
            method=qp_method,
            options=qp_options,
        )
        W, loss_W = res["x"], res["fun"]
        return W, loss_W.item()



import unittest
import pandas as pd

from pysyncon import Dataprep, AugSynth


class TestAugsynthBasque(unittest.TestCase):
    def setUp(self):
        df = pd.read_csv("./data/basque.csv")
        self.dataprep = Dataprep(
            foo=df,
            predictors=[
                "school.illit",
                "school.prim",
                "school.med",
                "school.high",
                "school.post.high",
                "invest",
            ],
            predictors_op="mean",
            time_predictors_prior=range(1964, 1970),
            special_predictors=[
                ("gdpcap", range(1960, 1970), "mean"),
                ("sec.agriculture", range(1961, 1970, 2), "mean"),
                ("sec.energy", range(1961, 1970, 2), "mean"),
                ("sec.industry", range(1961, 1970, 2), "mean"),
                ("sec.construction", range(1961, 1970, 2), "mean"),
                ("sec.services.venta", range(1961, 1970, 2), "mean"),
                ("sec.services.nonventa", range(1961, 1970, 2), "mean"),
                ("popdens", [1969], "mean"),
            ],
            dependent="gdpcap",
            unit_variable="regionname",
            time_variable="year",
            treatment_identifier="Basque Country (Pais Vasco)",
            controls_identifier=[
                "Andalucia",
                "Aragon",
                "Baleares (Islas)",
                "Canarias",
                "Cantabria",
                "Castilla-La Mancha",
                "Castilla Y Leon",
                "Cataluna",
                "Comunidad Valenciana",
                "Extremadura",
                "Galicia",
                "Madrid (Comunidad De)",
                "Murcia (Region de)",
                "Navarra (Comunidad Foral De)",
                "Principado De Asturias",
                "Rioja (La)",
                "Spain (Espana)",
            ],
            time_optimize_ssr=range(1960, 1970),
        )
        self.optim_method = "Nelder-Mead"
        self.optim_initial = "equal"
        self.weights = {
            "Andalucia": 0.113627911,
            "Aragon": 1.774922286,
            "Baleares (Islas)": -0.713432799,
            "Canarias": 1.19397534,
            "Cantabria": 0.497825351,
            "Castilla-La Mancha": 0.131573892,
            "Castilla Y Leon": -1.405974956,
            "Cataluna": 1.31890027,
            "Comunidad Valenciana": -1.731140541,
            "Extremadura": -1.134362989,
            "Galicia": 1.982136937,
            "Madrid (Comunidad De)": 0.110801212,
            "Murcia (Region de)": -1.31476635,
            "Navarra (Comunidad Foral De)": -1.303045915,
            "Principado De Asturias": -0.02423815,
            "Rioja (La)": 1.58950474,
            "Spain (Espana)": -0.086306241,
        }

    def test_weights(self):
        augsynth = AugSynth()
        augsynth.fit(dataprep=self.dataprep)

        weights = pd.Series(self.weights, name="weights")
        # Allow a tolerance of 2.5%
        pd.testing.assert_series_equal(
            weights, augsynth.weights(round=9), check_exact=False, atol=0.025
        )



import unittest
import numpy as np
import pandas as pd

from pysyncon.generator import LinearFactorModel


class TestLinearFactorModel(unittest.TestCase):
    def setUp(self):
        self.n_units = np.random.randint(low=10, high=20)
        self.n_observable = np.random.randint(low=10, high=15)
        self.n_unobservable = np.random.randint(low=10, high=15)
        self.n_periods_pre = np.random.randint(low=50, high=80)
        self.n_periods_post = np.random.randint(low=10, high=20)

    def test_matrix_dims(self):
        lfm = LinearFactorModel()
        X0, X1, Z0, Z1 = lfm.generate(
            n_units=self.n_units,
            n_observable=self.n_observable,
            n_unobservable=self.n_unobservable,
            n_periods_pre=self.n_periods_pre,
            n_periods_post=self.n_periods_post,
        )

        self.assertEqual(X0.shape, (self.n_observable, self.n_units - 1))
        self.assertEqual(X1.shape, (self.n_observable,))
        self.assertEqual(
            Z0.shape, (self.n_periods_pre + self.n_periods_post, self.n_units - 1)
        )
        self.assertEqual(Z1.shape, (self.n_periods_pre + self.n_periods_post,))



from __future__ import annotations
from typing import Optional

import numpy as np
import pandas as pd


class LinearFactorModel:
    """Generates potential outcomes following a linear factor model"""

    def __init__(
        self,
        observed_dist: tuple[int] = (0, 1),
        observed_params_dist: tuple[int] = (0, 10),
        unobserved_dist: tuple[int] = (0, 1),
        unobserved_params_dist: tuple[int] = (0, 10),
        effect_dist: tuple[int] = (0, 20),
        shocks_dist: tuple[int] = (0, 1),
        seed: Optional[int] = None,
        rng: Optional[np.random.Generator] = None,
    ) -> None:
        """Generates potential outcomes following a linear factor model

        Parameters
        ----------
        observed_dist : tuple, optional
            Parameters for the uniform distribution that the observed
            covariates follow, by default (0, 1)
        observed_params_dist : tuple, optional
            Parameters for the uniform distribution that the observed
            covariates model parameters follow, by default (0, 10)
        unobserved_dist : tuple, optional
            Parameters for the uniform distribution that the unobserved
            covariates follow, by default (0, 1)
        unobserved_params_dist : tuple, optional
            Parameters for the uniform distribution that the unobserved
            covariates model parameters follow, by default (0, 10)
        effect_dist : tuple, optional
            Uniform distribution parameters that the effect follows,
            by default (0, 20)
        shocks_dist : tuple, optional
            Normal distribution parameters that the shocks follow, by default (0, 1)
        seed : int, optional
            Random number generator seed, by default None
        rng : numpy.random.Generator, optional
            Provide a numpy random number generator, by default None
        """
        self.observed_dist = observed_dist
        self.observed_params_dist = observed_params_dist
        self.unobserved_dist = unobserved_dist
        self.unobserved_params_dist = unobserved_params_dist
        self.effect_dist = effect_dist
        self.shocks_dist = shocks_dist
        self.seed = seed
        self.rng = rng

    def generate(
        self,
        n_units: int,
        n_observable: int,
        n_unobservable: int,
        n_periods_pre: int,
        n_periods_post: int,
    ) -> tuple[pd.DataFrame, pd.Series, pd.DataFrame, pd.Series]:
        """Generate the matrices (:math:`X_0`, :math:`X_1`, :math:`Z_0`,
        :math:`Z_1`) that can be used as input to a synthetic control
        method (using the notation of Abadie & Gardeazabal :cite:`basque2003`).

        Parameters
        ----------
        n_units : int
            Number of units in the model
        n_observable : int
            Number of observable covariates in the model
        n_unobservable : int
            Number of unobservable covariates in the model
        n_periods_pre : int
            Number of time periods prior to the intervention
        n_periods_post : int
            Number of time periods post the intervention

        Returns
        -------
        tuple[pandas.DataFrame, pandas.Series, pandas.DataFrame, pandas.Series]
            Returns a tuple of 4 pandas objects: :math:`X_0` a pandas DataFrame
            of shape (n_periods_pre + n_periods_post, n_units - 1), :math:`X_1` a
            pandas Series of shape (n_periods_pre + n_periods_post, 1), :math:`Z_0`
            a pandas DataFrame of shape (n_observable, n_units - 1), :math:`Z_1`
            a pandas Series of shape (n_observable, 1).
        """
        rng = self.rng(self.seed) if self.rng else np.random.default_rng(seed=self.seed)

        n_periods = n_periods_pre + n_periods_post

        delta = rng.uniform(*self.effect_dist, size=n_periods).reshape(-1, 1)
        delta = np.column_stack([delta] * n_units)

        Z = rng.uniform(*self.observed_dist, size=(n_observable, n_units))
        mu = rng.uniform(*self.unobserved_dist, size=(n_unobservable, n_units))
        theta = rng.uniform(*self.observed_params_dist, size=(n_observable, n_periods))
        lambda_ = rng.uniform(
            *self.unobserved_params_dist, size=(n_unobservable, n_periods)
        )
        epsilon = rng.normal(*self.shocks_dist, size=(n_periods, n_units))

        Y_N = theta.T @ Z + lambda_.T @ mu + epsilon
        Y_I = Y_N + delta

        X0 = pd.DataFrame(
            data=Z[:, 1:],
            columns=range(2, n_units + 1),
            index=[f"observable{i}" for i in range(1, n_observable + 1)],
        )
        X1 = pd.Series(
            data=Z[:, 0],
            name=1,
            index=[f"observable{i}" for i in range(1, n_observable + 1)],
        )
        Z0 = pd.DataFrame(
            data=Y_N[:, 1:],
            columns=range(2, n_units + 1),
            index=range(1, n_periods + 1),
        )
        Z1 = pd.Series(
            data=np.concatenate(
                [Y_N[:n_periods_pre, 0], Y_I[n_periods_pre:, 0]], axis=0
            ),
            name=1,
            index=range(1, n_periods + 1),
        )
        return X0, X1, Z0, Z1



import unittest
import pandas as pd

from pysyncon import Dataprep, Synth
from pysyncon.utils import PlaceboTest


class TestSynthBasque(unittest.TestCase):
    def setUp(self):
        df = pd.read_csv("./data/basque.csv")
        self.dataprep = Dataprep(
            foo=df,
            predictors=[
                "school.illit",
                "school.prim",
                "school.med",
                "school.high",
                "school.post.high",
                "invest",
            ],
            predictors_op="mean",
            time_predictors_prior=range(1964, 1970),
            special_predictors=[
                ("gdpcap", range(1960, 1970), "mean"),
                ("sec.agriculture", range(1961, 1970, 2), "mean"),
                ("sec.energy", range(1961, 1970, 2), "mean"),
                ("sec.industry", range(1961, 1970, 2), "mean"),
                ("sec.construction", range(1961, 1970, 2), "mean"),
                ("sec.services.venta", range(1961, 1970, 2), "mean"),
                ("sec.services.nonventa", range(1961, 1970, 2), "mean"),
                ("popdens", [1969], "mean"),
            ],
            dependent="gdpcap",
            unit_variable="regionname",
            time_variable="year",
            treatment_identifier="Basque Country (Pais Vasco)",
            controls_identifier=[
                "Spain (Espana)",
                "Andalucia",
                "Aragon",
                "Principado De Asturias",
                "Baleares (Islas)",
                "Canarias",
                "Cantabria",
                "Castilla Y Leon",
                "Castilla-La Mancha",
                "Cataluna",
                "Comunidad Valenciana",
                "Extremadura",
                "Galicia",
                "Madrid (Comunidad De)",
                "Murcia (Region de)",
                "Navarra (Comunidad Foral De)",
                "Rioja (La)",
            ],
            time_optimize_ssr=range(1960, 1970),
        )
        self.optim_method = "Nelder-Mead"
        self.optim_initial = "equal"
        self.weights = {
            "Spain (Espana)": 0.0,
            "Andalucia": 0.0,
            "Aragon": 0.0,
            "Principado De Asturias": 0.0,
            "Baleares (Islas)": 0.0,
            "Canarias": 0.0,
            "Cantabria": 0.0,
            "Castilla Y Leon": 0.0,
            "Castilla-La Mancha": 0.0,
            "Cataluna": 0.850816306,
            "Comunidad Valenciana": 0.0,
            "Extremadura": 0.0,
            "Galicia": 0.0,
            "Madrid (Comunidad De)": 0.149183694,
            "Murcia (Region de)": 0.0,
            "Navarra (Comunidad Foral De)": 0.0,
            "Rioja (La)": 0.0,
        }
        self.placebo_gaps = {
            "Cataluna": {
                1960.0: 0.203808058,
                1961.0: 0.22013128,
                1962.0: 0.263867425,
                1963.0: 0.305086227,
                1964.0: 0.307812892,
                1965.0: 0.310500949,
                1966.0: 0.369694004,
                1967.0: 0.423575362,
                1968.0: 0.458736716,
                1969.0: 0.488697369,
                1970.0: 0.492355223,
            },
            "Madrid (Comunidad De)": {
                1960.0: 0.927170193,
                1961.0: 1.066511653,
                1962.0: 1.011029922,
                1963.0: 0.950455684,
                1964.0: 0.945846094,
                1965.0: 0.930053083,
                1966.0: 0.772220243,
                1967.0: 0.614648344,
                1968.0: 0.557832902,
                1969.0: 0.491439776,
                1970.0: 0.441262212,
            },
            "Andalucia": {
                1960.0: -0.005071144,
                1961.0: 0.002029757,
                1962.0: -0.002976465,
                1963.0: -0.008368432,
                1964.0: -0.012947738,
                1965.0: -0.018273511,
                1966.0: -0.002324632,
                1967.0: 0.012943551,
                1968.0: 0.009046557,
                1969.0: 0.004579814,
                1970.0: 0.013673678,
            },
        }
        self.summary = pd.DataFrame(
            data=[
                [7.26559110e-02, 3.98884646e01, 2.56336977e02, 3.23825543e02],
                [1.19777358e-01, 1.03174230e03, 2.73010720e03, 2.18245335e03],
                [3.48611100e-03, 9.03586680e01, 2.23340172e02, 1.48864075e02],
                [1.02189247e-01, 2.57275251e01, 6.34368045e01, 4.71326627e01],
                [1.08267860e-02, 1.34797198e01, 3.61534897e01, 2.61630325e01],
                [5.32110000e-05, 2.46473831e01, 2.15826359e01, 2.14454579e01],
                [1.17260969e-01, 5.28546845e00, 5.27078346e00, 3.58401509e00],
                [6.33926060e-02, 6.84399996e00, 6.17934020e00, 2.10581177e01],
                [1.55350772e-01, 4.10600004e00, 2.75975796e00, 5.25223529e00],
                [9.58688000e-02, 4.50820000e01, 3.76359420e01, 2.26702353e01],
                [5.30811070e-02, 6.15000000e00, 6.95245150e00, 7.27400001e00],
                [1.63475200e-03, 3.37540001e01, 4.11037607e01, 3.66458824e01],
                [2.37097130e-02, 4.07200012e00, 5.37134427e00, 7.10294116e00],
                [1.80712657e-01, 2.46889999e02, 1.96283316e02, 9.74682350e01],
            ],
            columns=["V", "treated", "synthetic", "sample mean"],
            index=[
                "school.illit",
                "school.prim",
                "school.med",
                "school.high",
                "school.post.high",
                "invest",
                "special.1.gdpcap",
                "special.2.sec.agriculture",
                "special.3.sec.energy",
                "special.4.sec.industry",
                "special.5.sec.construction",
                "special.6.sec.services.venta",
                "special.7.sec.services.nonventa",
                "special.8.popdens",
            ],
        )
        self.treatment_time = 1975
        self.pvalue = 0.16666666666666666
        self.att = {"att": -0.6995647842110987, "se": 0.07078092130438395}
        self.att_time_period = range(1975, 1998)
        self.mspe = 0.008864544955047298
        self.mape = 0.016928135318837897
        self.mae = 0.08777554288632104

    def test_weights(self):
        synth = Synth()
        synth.fit(
            dataprep=self.dataprep,
            optim_method=self.optim_method,
            optim_initial=self.optim_initial,
        )
        weights = pd.Series(self.weights, name="weights")
        # Allow a tolerance of 2.5%
        pd.testing.assert_series_equal(
            weights, synth.weights(round=9), check_exact=False, atol=0.025
        )
        pd.testing.assert_frame_equal(
            self.summary, synth.summary(round=9), check_exact=False, atol=0.025
        )

    def test_placebo_weights(self):
        synth = Synth()
        placebo_test = PlaceboTest()
        placebo_test.fit(
            dataprep=self.dataprep,
            scm=synth,
            scm_options={
                "optim_method": self.optim_method,
                "optim_initial": self.optim_initial,
            },
        )

        placebo_gaps = pd.DataFrame.from_dict(self.placebo_gaps).rename_axis(
            index="year"
        )
        regions = self.placebo_gaps.keys()
        years = list(self.placebo_gaps["Cataluna"].keys())
        pd.testing.assert_frame_equal(
            placebo_gaps,
            placebo_test.gaps[regions].loc[years],
            check_exact=False,
            atol=0.025,
        )
        self.assertAlmostEqual(
            self.pvalue,
            placebo_test.pvalue(treatment_time=self.treatment_time),
            places=3,
        )

    def test_att(self):
        synth = Synth()
        synth.fit(
            dataprep=self.dataprep,
            optim_method=self.optim_method,
            optim_initial=self.optim_initial,
        )
        synth_att = synth.att(time_period=self.att_time_period)

        # Allow a tolerance of 2.5%
        att_perc_delta = abs(1.0 - self.att["att"] / synth_att["att"])
        self.assertLessEqual(att_perc_delta, 0.025)

        # Allow a tolerance of 2.5%
        se_perc_delta = abs(1.0 - self.att["se"] / synth_att["se"])
        self.assertLessEqual(se_perc_delta, 0.025)

    def test_metric_values(self):
        synth = Synth()
        synth.fit(
            dataprep=self.dataprep,
            optim_method=self.optim_method,
            optim_initial=self.optim_initial,
        )

        # Allow a tolerance of 2.5%
        mspe_perc_delta = abs(1.0 - self.mspe / synth.mspe())
        self.assertLessEqual(mspe_perc_delta, 0.025)
        mape_perc_delta = abs(1.0 - self.mape / synth.mape())
        self.assertLessEqual(mape_perc_delta, 0.025)
        mae_perc_delta = abs(1.0 - self.mae / synth.mae())
        self.assertLessEqual(mae_perc_delta, 0.025)



import unittest
import pandas as pd

from pysyncon import Dataprep, Synth


class TestSynthGermany(unittest.TestCase):
    def setUp(self):
        df = pd.read_csv("./data/germany.csv")
        dataprep_train = Dataprep(
            foo=df,
            predictors=["gdp", "trade", "infrate"],
            predictors_op="mean",
            time_predictors_prior=range(1971, 1981),
            special_predictors=[
                ("industry", range(1971, 1981), "mean"),
                ("schooling", [1970, 1975], "mean"),
                ("invest70", [1980], "mean"),
            ],
            dependent="gdp",
            unit_variable="country",
            time_variable="year",
            treatment_identifier="West Germany",
            controls_identifier=[
                "USA",
                "UK",
                "Austria",
                "Belgium",
                "Denmark",
                "France",
                "Italy",
                "Netherlands",
                "Norway",
                "Switzerland",
                "Japan",
                "Greece",
                "Portugal",
                "Spain",
                "Australia",
                "New Zealand",
            ],
            time_optimize_ssr=range(1981, 1991),
        )
        synth_train = Synth()
        synth_train.fit(
            dataprep=dataprep_train, optim_method="Nelder-Mead", optim_initial="equal"
        )
        self.custom_V = synth_train.V

        self.dataprep = Dataprep(
            foo=df,
            predictors=["gdp", "trade", "infrate"],
            predictors_op="mean",
            time_predictors_prior=range(1981, 1991),
            special_predictors=[
                ("industry", range(1981, 1991), "mean"),
                ("schooling", [1980, 1985], "mean"),
                ("invest80", [1980], "mean"),
            ],
            dependent="gdp",
            unit_variable="country",
            time_variable="year",
            treatment_identifier="West Germany",
            controls_identifier=[
                "USA",
                "UK",
                "Austria",
                "Belgium",
                "Denmark",
                "France",
                "Italy",
                "Netherlands",
                "Norway",
                "Switzerland",
                "Japan",
                "Greece",
                "Portugal",
                "Spain",
                "Australia",
                "New Zealand",
            ],
            time_optimize_ssr=range(1960, 1990),
        )

        self.optim_method = "Nelder-Mead"
        self.optim_initial = "equal"
        self.weights = {
            "USA": 0.21624982,
            "UK": 0.0,
            "Austria": 0.414522077,
            "Belgium": 0.0,
            "Denmark": 0.0,
            "France": 0.0,
            "Italy": 0.0,
            "Netherlands": 0.09841208,
            "Norway": 0.0,
            "Switzerland": 0.107654851,
            "Japan": 0.163161172,
            "Greece": 0.0,
            "Portugal": 0.0,
            "Spain": 0.0,
            "Australia": 0.0,
            "New Zealand": 0.0,
        }
        self.att = {"att": -1555.1346777620479, "se": 317.6469306023242}
        self.att_time_period = range(1990, 2004)
        self.cis = {
            "value": {
                1991: 279.09685975333196,
                1992: 99.76203427529981,
                1993: -631.5437231770848,
                1994: -1050.2679900905205,
                1995: -1205.2549226793199,
                1996: -1467.2491625958974,
                1997: -1954.3741689815615,
                1998: -2008.3960300490326,
                1999: -2160.627036515649,
                2000: -2620.7330909274606,
            },
            "lower_ci": {
                1991: 43.148688105431994,
                1992: -136.18613737260014,
                1993: -867.4918948249846,
                1994: -1286.2161617384206,
                1995: -1441.20309432722,
                1996: -1703.1973342437975,
                1997: -2190.3223406294615,
                1998: -2244.3442016969325,
                1999: -2396.5752081635487,
                2000: -2856.6812625753605,
            },
            "upper_ci": {
                1991: 515.0450314012319,
                1992: 335.7102059231998,
                1993: -395.59555152918483,
                1994: -814.3198184426207,
                1995: -969.3067510314198,
                1996: -1231.3009909479972,
                1997: -1718.4259973336614,
                1998: -1772.4478584011324,
                1999: -1924.6788648677486,
                2000: -2384.7849192795607,
            },
        }
        self.ci_args = {
            "alpha": 0.05,
            "time_periods": [
                1991,
                1992,
                1993,
                1994,
                1995,
                1996,
                1997,
                1998,
                1999,
                2000,
            ],
            "max_iter": 50,
            "tol": 0.1,
            "verbose": False,
        }

    def test_weights(self):
        synth = Synth()
        synth.fit(
            dataprep=self.dataprep,
            optim_method=self.optim_method,
            optim_initial=self.optim_initial,
            custom_V=self.custom_V,
        )
        weights = pd.Series(self.weights, name="weights")
        pd.testing.assert_series_equal(
            weights, synth.weights(round=9), check_exact=False, atol=0.025
        )

    def test_att(self):
        synth = Synth()
        synth.fit(
            dataprep=self.dataprep,
            optim_method=self.optim_method,
            optim_initial=self.optim_initial,
            custom_V=self.custom_V,
        )
        synth_att = synth.att(time_period=self.att_time_period)

        # Allow a tolerance of 2.5%
        att_perc_delta = abs(1.0 - self.att["att"] / synth_att["att"])
        self.assertLessEqual(att_perc_delta, 0.025)

        # Allow a tolerance of 2.5%
        se_perc_delta = abs(1.0 - self.att["se"] / synth_att["se"])
        self.assertLessEqual(se_perc_delta, 0.025)

    def test_cis(self):
        synth = Synth()
        synth.fit(
            dataprep=self.dataprep,
            optim_method=self.optim_method,
            optim_initial=self.optim_initial,
            custom_V=self.custom_V,
        )

        cis = pd.DataFrame.from_dict(self.cis)
        cis.index.name = "time"
        pd.testing.assert_frame_equal(
            cis,
            synth.confidence_interval(custom_V=self.custom_V, **self.ci_args),
            check_exact=False,
            atol=0.025,
        )
